"""Dataset Descriptor"""

from typing import Dict, List
# unused imports
# from typing import Tuple
# import numpy as np

from .compartment_constraints import CompartmentConstraint
from .dataset import Dataset


class DatasetDescriptor:
    """Represents a dataset for fitting"""

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments
    # pylint: disable=attribute-defined-outside-init
    # Datasets are complex.

    def __init__(self,
                 label: str,
                 initial_concentration: str,
                 megacomplexes: List[str],
                 megacomplex_scaling: Dict[str, List[str]],
                 scaling: str,
                 compartment_scaling: Dict[str, List[str]],
                 compartment_constraints: List[CompartmentConstraint]):
        """

        Parameters
        ----------
        label : str
            The label of the dataset.

        initial_concentration : str
            The label of the initial concentration

        megacomplexes : List[str]
            A list of megacomplex labels

        megacomplex_scaling : Dict[str: List[str]]
            The megacomplex scaling parameters

        scaling : str
            The scaling parameter for the dataset

        compartment_scaling: Dict[str: List[str]]
            The compartment scaling parameters

        compartment_constraints: List[CompartmentConstraint] :
            A list of compartment constraints

        """
        self.label = label
        self.initial_concentration = initial_concentration
        self.megacomplexes = megacomplexes
        self.compartment_scaling = compartment_scaling
        self.megacomplex_scaling = megacomplex_scaling
        self.scaling = scaling
        self.dataset = None
        self.compartment_constraints = compartment_constraints

    @property
    def label(self):
        """The label of the dataset"""
        return self._label

    @label.setter
    def label(self, value):
        self._label = value

    @property
    def compartment_constraints(self):
        """A list of compartment constraints"""
        return self._compartment_constraints

    @compartment_constraints.setter
    def compartment_constraints(self, value):
        if not isinstance(value, list):
            value = [value]
        if any(not issubclass(type(val), CompartmentConstraint)
               for val in value):
            raise TypeError("CompartmentConstraint must be implementations of "
                            "model.CompartmentConstraint")
        self._compartment_constraints = value

    @property
    def dataset(self):
        """An implementation of model.Dataset"""
        return self._data

    @dataset.setter
    def dataset(self, data):
        if not isinstance(data, Dataset) and data is not None:
            raise TypeError
        self._data = data

    @property
    def initial_concentration(self):
        """A list of labels of initial concentrations"""
        return self._initial_concentration

    @initial_concentration.setter
    def initial_concentration(self, value):
        self._initial_concentration = value

    @property
    def scaling(self):
        """A parameter to scale the dataset"""
        return self._scaling

    @scaling.setter
    def scaling(self, scaling):
        if not isinstance(scaling, int) and scaling is not None:
            raise TypeError("Parameter index must be numerical")
        self._scaling = scaling

    @property
    def compartment_scaling(self):
        """ A dictionary of compartment parameter pairs"""
        return self._compartment_scaling

    @compartment_scaling.setter
    def compartment_scaling(self, scaling):
        if not isinstance(scaling, dict):
            raise TypeError
        self._compartment_scaling = scaling

    @property
    def megacomplexes(self):
        """list of megacomplex labels"""
        return self._megacomplexes

    @megacomplexes.setter
    def megacomplexes(self, megacomplex):
        if not isinstance(megacomplex, list):
            megacomplex = [megacomplex]
        if any(not isinstance(m, str) for m in megacomplex):
            raise TypeError("Megacomplex labels must be string.")
        self._megacomplexes = megacomplex

    @property
    def megacomplex_scaling(self):
        """A dictinary of megacomplex scaling parameters"""
        return self._megacomplex_scaling

    @megacomplex_scaling.setter
    def megacomplex_scaling(self, scaling):
        if not isinstance(scaling, dict):
            raise TypeError("Megacomplex Scaling must by dict, got"
                            "{}".format(type(scaling)))
        self._megacomplex_scaling = scaling

    def __str__(self):
        """ """
        string = "### _{}_\n\n".format(self.label)

        string += "* _Dataset Scaling_: {}\n".format(self.scaling)

        string += "* _Initial Concentration_: {}\n"\
            .format(self.initial_concentration)

        string += "* _Megacomplexes_: {}\n".format(self.megacomplexes)

        string += "* _Megacomplex Scalings_:"
        if self._megacomplex_scaling:
            string += "\n"
            for cmplx, scale in self._megacomplex_scaling.items():
                string += "  * {}:{}\n".format(cmplx, scale)
        else:
            string += " None\n"

        return string



class Megacomplex(object):
    """
    A megacomplex has a label.
    """
    def __init__(self, label):
        if not isinstance(label, str):
            raise TypeError
        self.label = label

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, label):
        self._label = label

    def __str__(self):
        return f"### _{self.label}_\n"



""" Glotaran DOAS Model Megacomplex """

from typing import List

from glotaran.models.spectral_temporal import KineticMegacomplex


class DOASMegacomplex(KineticMegacomplex):
    """Extends the glotaran.models.spectral_temporal.Megacomplex with
    oscillations"""
    def __init__(self, label, k_matrices: List[str], oscillations: List[str]):
        if not isinstance(oscillations, list):
            oscillations = [oscillations]
        self._oscillations = oscillations
        super(DOASMegacomplex, self).__init__(label, k_matrices)

    @property
    def oscillations(self):
        return self._oscillations

    @oscillations.setter
    def oscillations(self, value):
        if not isinstance(value, list):
            value = [value]
        if any(not isinstance(m, str) for m in value):
            raise TypeError
        self._oscillations = value

    def __str__(self):
        string = super(DOASMegacomplex, self).__str__()
        string += f"* _Oscillations_: {self.oscillations}\n"
        return string



import sys

import cython
cimport cython

import numpy as np
cimport numpy as np

# Not sure why but the scipy.special.cython_special.erfcx function doesnt' sseem to work.
#from scipy.special import erf, erfc, erfcx
#from scipy.special.cython_special cimport erf, erfc, erfcx #try in 0.19.0
#cimport scipy.special.cython_special as csc

from libc.math cimport exp,  pow, log, sqrt, erf
from numpy.math cimport NAN

cdef extern from "erfce.c":
    double erfce(double x)

def __init__():
    np.import_array()

@cython.boundscheck(False)
@cython.wraparound(False)
def calculateCSingleGaussian(double[:, :] C, idxs, double[:] k, double[:] x, double mu, double
                  delta, double scale, int backsweep, double backsweep_period):
    nr_x = x.shape[0]
    nr_comps = k.shape[0]
    cdef int n_c, n_x, n_k
    cdef double x_n, k_n, thresh, alpha, beta#term_1, term_2
    for n_k in range(nr_comps):
        k_n = k[n_k]
        if k_n == 0:
            continue
        n_c = idxs[n_k]
        alpha = (k_n * delta) / sqrt(2)
        for n_x in range(nr_x):
            x_n = x[n_x]
            beta = (x_n - mu) / (delta * sqrt(2))
            thresh = beta - alpha
            if thresh < -1 :
                C[n_x, n_c] += scale * .5 * erfce(-thresh) * exp(-beta * beta)
            else:
                C[n_x, n_c] += scale * .5 * (1 + erf(thresh)) * exp(alpha * (alpha - 2 * beta))
            if backsweep != 0:
                x1 = exp(-k_n * (x_n - mu + backsweep_period))
                x2 = exp(-k_n * ((backsweep_period / 2) - (x_n - mu)))
                x3 = exp(-k_n * backsweep_period)
                C[n_x, n_c] += (x1 + x2) / (1 - x3)


"""Glotaran DOAS Matrix"""

from typing import List
import numpy as np
from scipy.special import erf

from glotaran.model import Model, ParameterGroup
from glotaran.models.spectral_temporal import KineticMatrix


class DOASMatrix(KineticMatrix):
    """Extension of glotaran.models.spectral_temporal.KineticMatrix for a DOAS model."""
    def __init__(self, x: float, dataset: str, model: Model):
        """

        Parameters
        ----------
        x : float
            Point on the estimated axis the matrix calculated for

        dataset : str
            Dataset label of the dataset the matrix is calculated for

        model : glotaran.Model
            The model the matrix is calculated for


        """
        super(DOASMatrix, self).__init__(x, dataset, model)
        self._oscillations = []
        self._collect_oscillations(model)

    @property
    def compartment_order(self):
        """Sets the compartment order to map compartment labels to indices in
        the matrix"""
        compartment_order = super(DOASMatrix, self).compartment_order
        for osc in self._oscillations:
            compartment_order.append(osc.sin_compartment)
            compartment_order.append(osc.cos_compartment)
        return compartment_order

    def _collect_oscillations(self, model):
        for cmplx in [model.megacomplexes[mc] for mc in self.dataset.megacomplexes]:
            for label in cmplx.oscillations:
                self._oscillations.append(model.oscillations[label])

    def calculate(self,
                  matrix: np.array,
                  compartment_order: List[str],
                  parameter: ParameterGroup):
        """ Calculates the matrix.

        Parameters
        ----------
        matrix : np.array
            The preallocated matrix.

        compartment_order : list(str)
            A list of compartment labels to map compartments to indices in the
            matrix.

        parameter : lmfit.Parameters
            A dictory of parameters.


        Returns
        ^
        -------

        """
        super(DOASMatrix, self).calculate(matrix, compartment_order, parameter)

        idx = []
        freq = []
        rate = []
        for osc in self._oscillations:
            idx = compartment_order.index(osc.sin_compartment)
            freq = parameter.get(osc.frequency).value
            rate = parameter.get(osc.rate).value
            scale = self._dataset_scaling(parameter)
            osc = None
            if self._irf is None:
                osc = scale * np.exp(-rate * self.time - 1j * freq * self.time)
            else:
                centers, width, irf_scale, backsweep, backsweep_period = \
                        self._calculate_irf_parameter(parameter)
                d = width * width
                k = (rate + 1j * freq)

                a = np.exp((-1 * self.time + 0.5 * d * k) * k)
                b = 1 + erf((self.time - d * k) / (np.sqrt(2) * width))
                osc = a * b
            matrix[:, idx] = osc.real
            matrix[:, idx + 1] = osc.imag



from glotaran.model import Dataset
from glotaran.models.spectral_temporal.dataset import SpectralTemporalDataset
from .spectral_timetrace import SpectralUnit
# this import was unused and flake8 did complain, will leave it as comment
# from .spectral_timetrace import SpectralTimetrace
from enum import Enum
import os.path
import csv
import re
import numpy as np


class DataFileType(Enum):
    time_explicit = "Time explicit"
    wavelength_explicit = "Wavelength explicit"
    # TODO: implement time_intervals


class DataRequired(Exception):
    pass


class ExplicitFile(object):
    """
    Abstract class representing either a time- or wavelength-explicit file.
    """
    def __init__(self, filepath, *args, **kwargs):
        self._file_data_format = None
        self._observations = []  # TODO: choose name: data_points, observations, data
        self._times = []
        self._spectral_indices = []
        self._label = ""
        self._comment = ""
        absfilepath = os.path.realpath(filepath)
        if not filepath and 'dataset' in kwargs:
            self._initialize_with_dataset(kwargs.get('dataset'))
        elif os.path.isfile(filepath):
            self._file = filepath
        elif os.path.isfile(absfilepath):
            self._file = absfilepath
        else:
            raise Exception(f"Path does not exist: {filepath}, {absfilepath}")

    def get_explicit_axis(self):
        raise NotImplementedError

    def set_explicit_axis(self, axis):
        raise NotImplementedError

    def get_secondary_axis(self):
        raise NotImplementedError

    def get_data_row(self, index):
        raise NotImplementedError

    def get_observations(self, index):
        raise NotImplementedError

    def get_format_name(self):
        raise NotImplementedError

    def write_old(self, filename, export_type, overwrite=False, comment=""):
        if not isinstance(type, DataFileType):
            raise TypeError("Export type not supported")

        # self._dataset = dataset

        comment = comment.splitlines()
        while len(comment) < 2:
            comment.append("")

        if os.path.isfile(filename) and not overwrite:
            raise Exception("File already exist.")

        f = open(filename, "w")

        f.write(comment[0])
        f.write(comment[1])

        f.write(self.get_format_name())

        f.write("Intervalnr {}".format(len(self.get_explicit_axis())))

        datawriter = csv.writer(f, delimiter='\t')

        datawriter.writerow(self.get_explicit_axis())

        for i in range(len(self.get_secondary_axis())):
            datawriter.writerow(self.get_data_row(i)
                                .prepend(self.get_secondary_axis()[i]))

        f.close()

    def write(self, filename, overwrite=False, comment="",
              file_format="Time explicit", number_format="%.10e"):
        # TODO: write a more elegant method

        if os.path.isfile(filename) and not overwrite:
            print('File {} already exists'.format(os.path.isfile(filename)))
            raise Exception("File already exist.")

        comments = "# Filename: " + self._file + "\n" + " ".join(self._comment.splitlines()) + "\n"

        if file_format == "Wavelength explicit":
            wav = '\t'.join([repr(num) for num in self._spectral_indices])
            header = comments + "Wavelength explicit\nIntervalnr {}" \
                                "".format(len(self._spectral_indices)) + "\n" + wav
            raw_data = np.vstack((self._times.T, self._observations)).T
        elif file_format == "Time explicit":
            tim = '\t'.join([repr(num) for num in self._times])
            header = comments + "Time explicit\nIntervalnr {}" \
                                "".format(len(self._times)) + "\n" + tim
            raw_data = np.vstack((self._spectral_indices.T, self._observations.T)).T
        else:
            raise NotImplementedError

        np.savetxt(filename, raw_data, fmt=number_format, delimiter='\t', newline='\n',
                   header=header, footer='', comments='')

    def read(self, label, spectral_unit=SpectralUnit.nm, time_unit="s"):
        if not os.path.isfile(self._file):
            raise Exception("File does not exist.")

        self._label = label
        with open(self._file) as f:
            f.readline()  # Read first line with comments (and discard for now)
            f.readline()  # Read second line with comments (and discard for now)
            # TODO: what to do with return: None?
            self._file_data_format = get_data_file_format(f.readline())
            # TODO: what to do with return: None?
            interval_nr = get_interval_number(f.readline().strip().lower())
            all_data = []
            line = f.readline()
            while line:
                all_data.append([float(i) for i in re.split("\s+|\t+|\s+\t+|\t+\s+|\u3000+",
                                                            line.strip())])
                # data_block = pd.read_csv(line, sep="\s+|\t+|\s+\t+|\t+\s+|\u3000+",
                #                          engine='python', header=None, index_col=False)
                # all_data.append(data_block.values())
                line = f.readline()
            all_data = np.asarray(all_data)

            interval_counter = -interval_nr
            explicit_axis = []
            secondary_axis = []
            observations = [[]]
            obs_idx = 0

            for item in [sl for sublist in all_data for sl in sublist]:
                if item != item:  # NaN was found
                    ValueError()
                if interval_counter < 0:
                    # print("explicit_axis {}: {}".format(interval_counter, item))
                    explicit_axis.append(item)
                elif interval_counter == 0:
                    # print("secondary_axis: {}".format(item))
                    secondary_axis.append(item)
                elif interval_counter < (interval_nr + 1):
                    observations[obs_idx].append(item)
                interval_counter += 1
                if interval_counter == (interval_nr + 1):
                    interval_counter = 0
                    obs_idx += 1
                    observations.append([])

            del observations[-1]

            if self._file_data_format == DataFileType.time_explicit:
                self._times = np.asarray(explicit_axis)
                self._spectral_indices = np.asarray(secondary_axis)
                self._observations = np.asarray(observations)

            elif self._file_data_format == DataFileType.wavelength_explicit:
                self._spectral_indices = np.asarray(explicit_axis)
                self._times = np.asarray(secondary_axis)
                self._observations = np.asarray(observations)

            else:
                NotImplementedError()
                pass

        return self.dataset()

    def dataset(self):
        if not self._file_data_format:
            dataset = Dataset(self._label)
            pass
        elif self._file_data_format == DataFileType.time_explicit:
            dataset = SpectralTemporalDataset(self._label)
            dataset.set_axis("spectral", self._spectral_indices)
            dataset.set_axis("time", self._times)
        elif self._file_data_format == DataFileType.time_explicit:
            dataset = SpectralTemporalDataset(self._label)
            dataset.set_axis("time", self._times)
            dataset.set_axis("spectral", self._spectral_indices)
        dataset.set(self._observations)
        return dataset

    def _initialize_with_dataset(self, dataset):
        if isinstance(dataset, SpectralTemporalDataset):
            self._times = dataset.get_axis("time")
            self._spectral_indices = dataset.get_axis("spectral")
            self._observations = dataset.data
            self._file = "SpectralTemporalDataset"


class WavelengthExplicitFile(ExplicitFile):
    """
    Represents a wavelength explicit file
    """

    def get_explicit_axis(self):
        return self._spectral_indices

    def get_secondary_axis(self):
        return self.observations()

    def get_data_row(self, index):
        row = []
        pass
        return row

    def add_data_row(self, row):
        if self._timepoints is None:
            self._timepoints = []
        self._timepoints.append(float(row.pop(0)))

        if self._spectra is None:
            self._spectra = []
        self._spectra.append(float(row))

    def get_format_name(self):
        return DataFileType.wavelength_explicit

    def times(self):
        return self.get_secondary_axis()

    def wavelengths(self):
        return self.get_explicit_axis()


class TimeExplicitFile(ExplicitFile):
    """
    Represents a time explicit file
    """
    def get_explicit_axis(self):
        return self.observations()

    def set_explicit_axis(self, axies):
        self._timepoints = float(axies)

    def get_secondary_axis(self):
        return self.channel_labels

    def get_data_row(self, index):
        return self.get_channel(self.channel_labels()[index])

    def add_data_row(self, row):
        if self._spectral_indices is None:
            self._spectral_indices = []
        self._spectral_indices.append(row.pop(0))

        if self._spectra is None:
            self._spectra = []
        self._spectra.append(float(row))

    def get_format_name(self):
        return DataFileType.time_explicit


def get_interval_number(line):
    interval_number = None
    match = re.search(r"intervalnr\s(.*)", line.strip().lower())
    if match:
        interval_number = match.group(1)
    if not interval_number:
        interval_number = re.search(r'\d+', line[::-1]).group()[::-1]
    try:
        interval_number = int(interval_number)
    except ValueError:
        pass  # TODO: let user know no interval_number was found
        interval_number = None
    return interval_number


def get_data_file_format(line):
    data_file_format = None
    if re.search(r"time\s+explicit|time\t+explicit", line.strip().lower()):
        # print("Time explicit format") #TODO: verbosity / debug statement
        data_file_format = DataFileType.time_explicit
    elif re.search(r"wavelength\s+explicit|wavelength\t+explicit", line.strip().lower()):
        # print("Wavelength explicit format") #TODO: verbosity / debug statement
        data_file_format = DataFileType.wavelength_explicit
    else:
        NotImplementedError()
    return data_file_format



"""Glotarans model package"""
from . import (
               compartment_constraints,
               dataset,
               dataset_descriptor,
               initial_concentration,
               megacomplex,
               model,
               parameter,
               parameter_constraints,
               parameter_group,
               )


# Compartment Constraints
CompartmentConstraintType = compartment_constraints.CompartmentConstraintType
CompartmentConstraint = compartment_constraints.CompartmentConstraint
ZeroConstraint = compartment_constraints.ZeroConstraint
EqualConstraint = compartment_constraints.EqualConstraint
EqualAreaConstraint = compartment_constraints.EqualAreaConstraint

# Dataset

DatasetDescriptor = dataset_descriptor.DatasetDescriptor
Dataset = dataset.Dataset

# Initial Concentration

InitialConcentration = initial_concentration.InitialConcentration

# Megacomplex

Megacomplex = megacomplex.Megacomplex

# Model

Model = model.Model

# Parameter

Parameter = parameter.Parameter
ParameterGroup = parameter_group.ParameterGroup

# Parameter Constraint

BoundConstraint = parameter_constraints.BoundConstraint
FixedConstraint = parameter_constraints.FixedConstraint



from itertools import chain


def get_keys_from_object(obj, keys, start=0):
    if is_compact(obj):
        idx = [i+start for i in range(len(keys))]
        return tuple(list([obj[i] if len(obj) > i else None for i in idx]))
    else:
        return tuple(list([obj[k] if k in obj else None for k in keys]))


def retrieve_optional(obj, key, index, default):
    compact = is_compact(obj)
    if compact:
        if len(obj) <= index:
            return default
        return obj[index]
    else:
        if key in obj:
            return obj[key]
        return default


def parse_range(rng):
    parts = rng.split('-')

    if 1 > len(parts) > 2:
        raise ValueError("Bad range: '%s'" % (rng,))
    start = parts[0]
    if len(parts) == 1:
        yield "{}".format(start).strip()
    else:
        path = []
        if not start.isdigit():
            path = start.split(".")
            start = path.pop()
        start = int(start)
        end = parts[1]
        if not end.isdigit():
            end = end.split(".").pop()
        end = int(end)
        if start > end:
            end, start = start, end
        if path:
            for i in range(start, end + 1):
                yield ".".join(path) + ".{}".format(i).strip()
        else:
            for i in range(start, end + 1):
                yield "{}".format(i).strip()


def parse_range_list(rngs):
    return sorted(set(chain(*[parse_range(rng) for rng in rngs.split(',')])))


def is_compact(obj):
    return isinstance(obj, list)



"""Glotaran Kinetic Model"""

from typing import Type, Dict
from glotaran.model import Model

from glotaran.fitmodel import FitModel, Matrix

from .dataset import SpectralTemporalDataset

from .irf import Irf
from .fitmodel import KineticFitModel
from .k_matrix import KMatrix
from .kinetic_matrix import KineticMatrix
from .megacomplex import KineticMegacomplex
from .spectral_matrix import SpectralMatrix
from .spectral_shape import SpectralShape


class KineticModel(Model):
    """A kinetic model is an implementation for model.Model. It is used describe
    time dependend datasets.

    """

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments
    # pylint: disable=attribute-defined-outside-init
    # Models are complex.
    def __init__(self):
        """ """
        self.k_matrices = {}
        self.irfs = {}
        self.shapes = {}
        super(KineticModel, self).__init__()

    def type_string(self) -> str:
        """Returns a human readable string identifying the type of the model.

        Returns
        -------

        type : str
            Type of the Model

        """
        return "Kinetic"

    def calculated_matrix(self) -> Type[Matrix]:
        """Returns Kinetic C the calculated matrix.

        Returns
        -------

        matrix : type(fitmodel.Matrix)
            Calculated Matrix
        """
        return KineticMatrix

    def estimated_matrix(self) -> Type[Matrix]:
        """Returns the estimated matrix.

        Returns
        -------

        matrix : np.array
            Estimated Matrix
        """
        return SpectralMatrix

    def fit_model_class(self) -> Type[FitModel]:
        """Returns a kinetic fitmodel.

        Returns
        -------

        fitmodel : type(KineticFitModel)
            Implementation of fitmodel.FitModel
        """
        return KineticFitModel

    def dataset_class(self) -> Type[SpectralTemporalDataset]:
        """Returns an implementation for model.Dataset

        Returns
        -------

        descriptor : type(model.Dataset)
            Implementation of model.Dataset
        """
        return SpectralTemporalDataset

    def add_megacomplex(self, megacomplex: KineticMegacomplex):
        """Adds a kinetic megacomplex to the model.

        Parameters
        ----------
        megacomplex : KineticMegacomplex


        """
        if not isinstance(megacomplex, KineticMegacomplex):
            raise TypeError
        super(KineticModel, self).add_megacomplex(megacomplex)

    @property
    def k_matrices(self) -> Dict[str, KMatrix]:
        """A dictonary of the models K-matrices"""
        return self._k_matrices

    @k_matrices.setter
    def k_matrices(self, value):
        if not isinstance(value, dict):
            raise TypeError("K-Matrices must be dict.")
        if any(not isinstance(type(val), KMatrix) for val in value):
            raise TypeError("K-Matrices must be subclass of 'KMatrix'")
        self._k_matrices = value

    def add_k_matrix(self, k_matrix: KMatrix):
        """Adds a K-Matrix to the model.

        Parameters
        ----------
        k_matrix : KMatrix


        """
        if not issubclass(type(k_matrix), KMatrix):
            raise TypeError("K-Matrix must be subclass of 'KMatrix'")
        if self.k_matrices is None:
            self.k_matrices = {k_matrix.label: k_matrix}
        else:
            if k_matrix.label in self.k_matrices:
                raise Exception("K-Matrix labels must be unique")
            self.k_matrices[k_matrix.label] = k_matrix

    @property
    def irfs(self) -> Dict[str, Irf]:
        """A dictonary of the model's IRFs"""
        return self._irfs

    @irfs.setter
    def irfs(self, value):
        if not isinstance(value, dict):
            raise TypeError("Irfs must be dict.")
        if any(issubclass(type(val), Irf) for val in value):
            raise TypeError("Irfs must be subclass of 'KMatrix'")
        self._irfs = value

    def add_irf(self, irf: Irf):
        """Adds an IRF to model

        Parameters
        ----------
        irf : Irf


        """
        if not issubclass(type(irf), Irf):
            raise TypeError("Irfs must be subclass of 'Irf'")
        if self.irfs is None:
            self.irfs = {irf.label: irf}
        else:
            if irf.label is None or irf.label == "":
                raise Exception("Irf label empty")
            if irf.label in self.irfs:
                raise Exception("Irf label must be unique")
            self.irfs[irf.label] = irf

    @property
    def shapes(self) -> Dict[str, SpectralShape]:
        """A Dictonary of the model's spectral shapes."""
        return self._shapes

    @shapes.setter
    def shapes(self, value):
        if not isinstance(value, dict):
            raise TypeError("Shapes must be dict.")
        if any(issubclass(type(val), SpectralShape) for val in value):
            raise TypeError("Values must be subclass of 'Shape'")
        self._shapes = value

    def add_shape(self, shape: SpectralShape):
        """Adds a spectral shape to the model.

        Parameters
        ----------
        shape : SpectralShape


        """
        if not issubclass(type(shape), SpectralShape):
            raise TypeError("Shape must be subclass of 'Shape'")
        if self.shapes is None:
            self.shapes = {shape.label: shape}
        else:
            if shape.label in self.shapes:
                raise Exception("Shape labels must be unique")
            self.shapes[shape.label] = shape

    def get_megacomplex_k_matrix(self, cmplx: str) -> KMatrix:
        cmplx = self.megacomplexes[cmplx]
        kmat = KMatrix.empty(cmplx.label, self.compartments)
        for mat in cmplx.k_matrices:
            kmat = kmat.combine(self.k_matrices[mat])
        return kmat

    def __str__(self):
        string = super(KineticModel, self).__str__()
        string += "\n## K-Matrices\n\n"
        for k in self.k_matrices:
            string += f"{self.k_matrices[k]}\n"

        if self.irfs:
            string += "## IRFs\n\n"
            for irf in self.irfs:
                string += f"{self.irfs[irf]}\n"

        if self.shapes:
            string += "## Shapes\n\n"
            for _, shape in self.shapes.items():
                string += f"{shape}\n"
        return string



"""Glotaran DOAS Model Oscillation"""


class Oscillation:
    """A damped oscillation"""
    def __init__(self, label: str, compartment: str, frequency: str, rate: str):
        """
        Parameter
        ---------
        frequency : str
            The frequency of the oscillation
        rate : str
            The damping rate of the oscillation
        """
        self._label = label
        self._compartment = compartment
        self._frequency = frequency
        self._rate = rate

    @property
    def label(self) -> str:
        """The label of the oscillation"""
        return self._label

    @property
    def compartment(self) -> str:
        """The the target compartment of the oscillation"""
        return self._compartment

    @property
    def sin_compartment(self) -> str:
        """The the sin target compartment of the oscillation"""
        return f"{self._compartment}_sin"

    @property
    def cos_compartment(self) -> str:
        """The the cos target compartment of the oscillation"""
        return f"{self._compartment}_cos"

    @property
    def frequency(self) -> str:
        """The frequency of the oscillation"""
        return self._frequency

    @property
    def rate(self) -> str:
        """The damping rate of the oscillation"""
        return self._rate

    def __str__(self):
        string = f"_{self.label}_:"
        string += f" __Compartment__: {self.compartment}"
        string += f" __Rate__: {self.rate}"
        string += f" __Frequency__: {self.frequency}"
        return string



from glotaran.models.spectral_temporal.dataset import SpectralTemporalDataset
from natsort import natsorted
from glob import glob
import os
import numpy as np


def load_samples(samples_path):
    with open(samples_path, 'rb') as f:
        shape = get_header(f, 2)
        data = get_data(f, shape)
        return data


def load_times(times_path):
    with open(times_path, 'rb') as f:
        shape = get_header(f, 1)
        data = get_data(f, shape)
        return data


def get_header(f, dims):
    header = np.fromfile(f, dtype=">i", count=dims)
    return header


def get_data(f, shape):
    data = np.fromfile(f, dtype=">d")
    return data.reshape(shape).astype(np.float)


class ChlorospecData(object):
    """
    Class capable of reading in binary data in the structure
    FOLDER/NEXPERIMENT/NREPEAT/DATA
    Returns a glotaran.model.dataset
    """

    def __init__(self, folder, debug=False):
        self._data_folder = folder
        self._label = ""

    def read(self, label):
        if not os.path.isdir(self._data_folder):
            raise Exception("Data folder does not exist.")
        if not ChlorospecData.is_valid_path(self._data_folder):
            raise Exception("Not a valid data folder")

        # with open(self._data_folder) as f:
        # Recognize binary signature?
        # self._file_data_format = get_data_file_format(f)  # TODO: what to do with return: None?
        # if not self._file_data_format:
        #    ImportError

        self._label = label
        f = self._data_folder
        sorted_sub_folders = ChlorospecData.valid_sub_folders_natural_sorted(f)
        t0 = 0
        all_times = np.empty(shape=0)
        data = np.empty(shape=(0, 0))

        for idx, s in enumerate(sorted_sub_folders):
            sf = os.path.join(f, s)
            times_path = os.path.join(sf, "times.bin")
            samples_path = os.path.join(sf, "spectra.bin")
            if os.path.isfile(times_path) and os.path.isfile(samples_path):
                times = load_times(times_path)
                if idx == 0:
                    t0 = times[0]
                times = [-t0 + t for t in times]
                all_times = np.concatenate((all_times, times))
                samples = load_samples(samples_path)
                if data.size == 0:
                    data = samples.T  # our dataset convention
                else:
                    data = np.concatenate((data, samples.T), 1)
        wavelengths = np.linspace(159.735, 1047.157, data.shape[0])
        dataset = SpectralTemporalDataset(self._label)
        dataset.set_axis("time", all_times)
        dataset.set_axis("spectral", wavelengths)
        dataset.data = data
        return dataset

    @staticmethod
    def valid_sub_folders_natural_sorted(path):
        sub_folders = [f.path for f in os.scandir(path) if f.is_dir() and
                       glob(os.path.join(f.path, '*.bin'))]
        return natsorted(sub_folders, key=lambda x: x.lower())

    @staticmethod
    def is_valid_path(path):
        return os.path.isdir(path) and \
               len(ChlorospecData.valid_sub_folders_natural_sorted(path)) > 0



class Irf(object):
    """Represents an IRF."""

    def __init__(self, label):
        self.label = label

    @property
    def label(self):
        """Label of the IRF"""
        return self._label

    @label.setter
    def label(self, value):
        """

        Parameters
        ----------
        value : label of the IRF


        Returns
        -------


        """
        if not isinstance(value, str):
            raise TypeError("Labels must be strings.")
        self._label = value

    def type_string(self):
        """Identifies an implementation """
        raise NotImplementedError

    def __str__(self):
        return f"### _{self.label}_\n* _Type_: {self.type_string()}\n"



"""Glotaran Fitmodel Matrix Group"""

from typing import List, Tuple
import collections
import numpy as np

from glotaran.model.dataset_descriptor import DatasetDescriptor
from glotaran.model.parameter_group import ParameterGroup

from .matrix import Matrix


class MatrixGroup:
    """A group of matrices at an index on the estimated or calculated axis. The
    matrices are concatinated along estimated or calculated axis.

    Parameters
    ----------

    Returns
    -------

    """
    def __init__(self, matrix: Matrix):
        """

        Parameters
        ----------
        matrix: Matrix
            The initial matrix

        """
        self.matrices = [matrix]

    def add_matrix(self, matrix: Matrix):
        """Adds matrix to the group.

        Parameters
        ----------
        matrix : fitmodel.Matrix
            Matrix to be added.

        """
        self.matrices.append(matrix)

    def calculate(self, parameter: ParameterGroup) -> np.array:
        """Calculates the matrix group.

        Parameters
        ----------
        parameter : ParameterGroup
            Parameters for the calculation

        Returns
        -------
        matrix : numpy.array
            Calculated matrix


        """

        compartment_order = self.compartment_order()

        matrix = np.zeros(self.shape(), dtype=np.float64)

        t_idx = 0

        for mat in self.matrices:
            n_t = t_idx + mat.shape[0]
            mat.calculate(matrix[t_idx:n_t, :], compartment_order,
                          parameter)
            mat.apply_constraints(matrix[t_idx:n_t, :], compartment_order,
                                  parameter)
            t_idx = n_t

        return matrix

    def matrix_location(self, dataset: DatasetDescriptor) -> Tuple[int, int]:
        """Returns the location of a dataset's matrix within the matrix group

        Parameters
        ----------
        dataset : DatasetDescriptor
            Descriptor of the dataset

        Returns
        -------
        location : tuple(int, int)
            The location of the datset. None if dataset is not present.

        """
        t_idx = 0
        for mat in self.matrices:
            if mat.dataset.label == dataset.label:
                return (t_idx, mat.shape()[0])
            t_idx += mat.shape()[0]
        return None

    def sizes(self) -> List[int]:
        """Returns list with all sizes of the matrices in the concating
        direction.

        Returns
        -------
        sizes : list(int)

        """
        return [mat.shape[0] for mat in self.matrices]

    def compartment_order(self) -> List[str]:
        """A list of compartments which maps entries in the matrix to
        compartments in the model.

        Returns
        -------
        compartment_order : list(str)

        """
        compartment_order = [c for cmat in self.matrices
                             for c in cmat.compartment_order]
        return list(collections.OrderedDict.fromkeys(compartment_order).keys())

    def shape(self) -> Tuple[int, int]:
        """The shape of the whole matrix group.

        Returns
        -------
        shape : tuple(int, int)
        """
        dim0 = sum([mat.shape[0] for mat in self.matrices])
        dim1 = len(self.compartment_order())
        return (dim0, dim1)



""" Glotaran Initial Concentration"""

from typing import List


class InitialConcentration(object):
    """
    An initial concentration vector.
    """
    def __init__(self, label: str, parameter: List[str]):
        self.label = label
        self.parameter = parameter

    @property
    def label(self) -> str:
        """The label of the initial concentration"""
        return self._label

    @label.setter
    def label(self, value):
        self._label = value

    @property
    def parameter(self) -> List[str]:
        """A list of parameters representing the concentrations of the
        compartments."""
        return self._parameter

    @parameter.setter
    def parameter(self, value):
        if not isinstance(value, list):
            value = [value]
        self._parameter = value

    def __str__(self):
        return f"* __{self.label}__: {self.parameter}"



"""Glotaran DOAS Model Spectral Matrix"""

from glotaran.models.spectral_temporal import SpectralMatrix


class DOASSpectralMatrix(SpectralMatrix):

    def collect_shapes(self):
        super(DOASSpectralMatrix, self).collect_shapes()
        cmplxs = [self.model.megacomplexes[c] for c in self.dataset.megacomplexes]
        oscs = [self.model.oscillations[o] for cmplx in cmplxs
                for o in cmplx.oscillations]
        oscs = list(set(oscs))
        for osc in oscs:
            if osc in self.shapes:
                shape = self.shapes[osc]
                self.shapes[f"{osc}_sin"] = shape
                self.shapes[f"{osc}_cos"] = shape

    @property
    def compartment_order(self):
        """Sets the compartment order to map compartment labels to indices in
        the matrix"""
        compartment_order = super(DOASSpectralMatrix, self).compartment_order
        cmplxs = [self.model.megacomplexes[c] for c in self.dataset.megacomplexes]
        oscs = [self.model.oscillations[o] for cmplx in cmplxs
                for o in cmplx.oscillations]
        oscs = list(set(oscs))
        for osc in oscs:
            compartment_order.append(osc.sin_compartment)
            compartment_order.append(osc.cos_compartment)
        return compartment_order



"""Glotaran Kinetic Fitmodel"""
from glotaran.fitmodel import FitModel

from .result import KineticResult


class KineticFitModel(FitModel):
    """Thin wrapper to return KineticResult instead of fitmodel.Result"""
    def result_class(self) -> KineticResult:
        """Returns KineticResult"""
        return KineticResult



from . import (
    dataset_descriptor,
    irf_gaussian,
    irf_measured,
    k_matrix,
    kinetic_matrix,
    megacomplex,
    model_kinetic,
    spectral_matrix,
    spectral_shape_gaussian,
)

# Dataset Descriptor

SpectralTemporalDatasetDescriptor = \
    dataset_descriptor.SpectralTemporalDatasetDescriptor

# Irfs

GaussianIrf = irf_gaussian.GaussianIrf
MeasuredIrf = irf_measured.MeasuredIrf

# Shapes

SpectralShapeGaussian = spectral_shape_gaussian.SpectralShapeGaussian

# Matrix

KineticMatrix = kinetic_matrix.KineticMatrix
SpectralMatrix = spectral_matrix.SpectralMatrix

# K Matrix

KMatrix = k_matrix.KMatrix

# Megacomplex

KineticMegacomplex = megacomplex.KineticMegacomplex

# Model

KineticModel = model_kinetic.KineticModel



"""Glotaran Fitmodel"""

from typing import List, Type
import numpy as np
from lmfit_varpro import CompartmentEqualityConstraint, SeparableModel
from lmfit import Parameters

from glotaran.model.parameter_group import ParameterGroup

from .matrix_group_generator import MatrixGroupGenerator
from .result import Result


class FitModel(SeparableModel):
    """FitModel is an implementation of lmfit-varpor.SeparableModel."""

    # pylint: disable=no-self-use
    # pylint: disable=arguments-differ
    # we do some convinince wrapping

    def __init__(self, model: 'glotaran.Model'):
        """

        Parameters
        ----------
        model : glotaran.Model

        """
        self._model = model
        self._generator = None
        self._dataset_group = None

    @property
    def model(self):
        """The underlying glotaran.Model"""
        return self._model

    def get_initial_fitting_parameter(self) -> Parameters:
        """

        Returns
        -------
        Parameters : lmfit.Parameters
        """
        return self._model.parameter.as_parameters_dict(only_fit=True)

    def data(self, **kwargs) -> List[np.array]:
        """ Returns the data to fit.


        Returns
        -------
        data: list(np.array)
        """
        if "dataset" in kwargs:
            label = kwargs["dataset"]
            gen = MatrixGroupGenerator.for_dataset(self._model, label,
                                                   self._model.
                                                   calculated_matrix())
            return gen.create_dataset_group()
        return self._dataset_group

    def fit(self, *args, nnls=False, **kwargs) -> Result:
        """Fits the model.

        Parameters
        ----------
        nnls :
             (Default value = False)
             Use Non-Linear Least Squares instead of variable projection.
        *args :

        **kwargs :


        Returns
        -------
        result : Result

        """

        result = self.result(nnls, *args, **kwargs)

        result.fit(*args, **kwargs)
        return result

    def result_class(self) -> Type[Result]:
        """Returns a Result class implementation. Meant to be overwritten."""
        return Result

    def result(self, nnls: bool, *args, **kwargs) -> Result:
        """Creates a Result object.

        Parameters
        ----------
        nnls : bool
             Use Non-Linear Least Squares instead of variable projection.

        *args :

        **kwargs :


        Returns
        -------
        result : Result
        """
        self._generator = MatrixGroupGenerator.for_model(self._model,
                                                         self._model.
                                                         calculated_matrix())
        self._dataset_group = self._generator.create_dataset_group()
        c_constraints = self._create_constraints()
        result = self.result_class()(self,
                                     self.get_initial_fitting_parameter(),
                                     nnls,
                                     c_constraints,
                                     *args, **kwargs)
        return result

    def c_matrix(self, parameter: Parameters, *args, **kwargs) -> np.array:
        """Implementation of SeparableModel.c_matrix.

        Parameters
        ----------
        parameter : lmfit.Parameters

        *args :

        **kwargs :
            dataset : str
                Only evaluate for the given dataset


        Returns
        -------
        matrix : np.array
        """
        parameter = ParameterGroup.from_parameter_dict(parameter)
        if "dataset" in kwargs:
            label = kwargs["dataset"]
            gen = MatrixGroupGenerator.for_dataset(self._model, label,
                                                   self._model.
                                                   calculated_matrix())
        else:
            if self._generator is None:
                self._init_generator()
            gen = self._generator
        return gen.calculate(parameter)

    def get_calculated_matrix_group(self, dataset=None):
        if dataset is None:
            return MatrixGroupGenerator.for_model(self._model,
                                                  self._model.
                                                  calculated_matrix())

        return MatrixGroupGenerator.for_dataset(self._model, dataset,
                                                self._model.
                                                calculated_matrix())

    def _init_generator(self):
        self._generator = MatrixGroupGenerator.for_model(self._model,
                                                         self._model.
                                                         calculated_matrix())

    def e_matrix(self, parameter, *args, **kwargs) -> np.array:
        """Implementation of SeparableModel.e_matrix.

        Parameters
        ----------
        parameter : lmfit.Parameters

        *args :

        **kwargs :
            dataset : str
                Only evaluate for the given dataset
            axis : np.array
                The axis to evaluate the e-matrix on.


        Returns
        -------
        matrix : np.array
        """
        # We don't have a way to construct a complete E matrix for the full
        # problem yet.
        if "dataset" not in kwargs:
            raise Exception("'dataset' non specified in kwargs")

        parameter = ParameterGroup.from_parameter_dict(parameter)
        dataset = self._model.datasets[kwargs["dataset"]]

        # A data object needs to be present to provide axies
        if dataset.dataset is None:
            raise Exception("No Data object present for dataset '{}'"
                            .format(kwargs["dataset"]))

        axis = kwargs["axis"] if "axis" in kwargs else np.asarray([0])

        e_matrix = self._model.estimated_matrix()(axis, dataset, self._model)
        return e_matrix.calculate_standalone(parameter)

    def _create_constraints(self) -> List[CompartmentEqualityConstraint]:
        c_constraints = []
        for _, dataset in self._model.datasets.items():
            constraints = [c for c in dataset.compartment_constraints if
                           c.type() == 2]

            for cons in constraints:
                for interval in cons.intervals:
                    group = list(self._generator.groups_in_range(interval))[0]
                    crange = group.get_dataset_location(dataset)
                    i = group.compartment_order.index(cons.target)
                    j = group.compartment_order.index(cons.compartment)
                    c_constraints.append(
                        CompartmentEqualityConstraint(cons.weight,
                                                      i, j,
                                                      cons.parameter,
                                                      interval,
                                                      crange))
        return c_constraints


def isclass(obj, classname):
    """ Checks if an objects classname matches the given classname

    Parameters
    ----------
    obj : any

    classname : str


    Returns
    -------
    isclass : bool
    """
    return obj.__class__.__name__ == classname



""" Gloataran DOAS Model """

from typing import Dict, Type

from glotaran.fitmodel import Matrix
from glotaran.models.spectral_temporal import KineticModel

from .doas_matrix import DOASMatrix
from .doas_spectral_matrix import DOASSpectralMatrix
from .oscillation import Oscillation


class DOASModel(KineticModel):
    """Extends the kinetic model with damped oscillations."""
    def __init__(self):
        """ """
        self.oscillations = {}
        super(DOASModel, self).__init__()

    def type_string(self) -> str:
        """Returns a human readable string identifying the type of the model.

        Returns
        -------

        type : str
            Type of the Model

        """
        return "Kinetic"

    def calculated_matrix(self) -> Type[Matrix]:
        """Returns Kinetic C the calculated matrix.

        Returns
        -------

        matrix : type(fitmodel.Matrix)
            Calculated Matrix
        """
        return DOASMatrix

    def estimated_matrix(self) -> Type[Matrix]:
        """Returns the estimated matrix.

        Returns
        -------

        matrix : np.array
            Estimated Matrix
        """
        return DOASSpectralMatrix

    @property
    def oscillations(self) -> Dict[str, Oscillation]:
        """A dictonary of the models oscillations."""
        return self._oscillations

    @oscillations.setter
    def oscillations(self, value: Dict[str, Oscillation]):
        if not isinstance(value, dict):
            raise TypeError("Oscillations must be dict.")
        if any(not isinstance(type(val), Oscillation) for val in value):
            raise TypeError("Oscillations must be subclass of 'Oscillation'")
        self._oscillations = value

    def add_oscillation(self, oscillation):
        """

        Parameters
        ----------


        Returns
        -------

        """
        if not issubclass(type(oscillation), Oscillation):
            raise TypeError("Oscillation must be subclass of 'Oscillation")
        if self.oscillations is None:
            self.oscillations = {oscillation.label: oscillation}
        else:
            if oscillation.label in self.oscillations:
                raise Exception("Oscillation labels must be unique")
            self.oscillations[oscillation.label] = oscillation

    def __str__(self):
        string = super(DOASModel, self).__str__()
        string += "\n## Oscillations\n\n"
        for o in self.oscillations:
            string += f"{self.oscillations[o]}\n"

        if self.irfs:
            string += "## IRFs\n\n"
            for irf in self.irfs:
                string += f"{self.irfs[irf]}\n"

        if self.shapes:
            string += "## Shapes\n\n"
            for _, shape in self.shapes.items():
                string += f"{shape}\n"
        return string



from .irf import Irf


class GaussianIrf(Irf):
    """
    Represents a gaussian IRF.

    One width and one center is a single gauss.

    One center and multiple widths is a multiple gaussian.

    Multiple center and multiple widths is Double-, Triple- , etc. Gaussian.

    Parameters
    ----------

    label:
        label of the irf
    center:
        one or more center of the irf as parameter indices
    width:
        one or more widths of the gaussian as parameter index
    center_dispersion:
        polynomial coefficients for the dispersion of the
        center as list of parameter indices. None for no dispersion.
    width_dispersion:
        polynomial coefficients for the dispersion of the
        width as parameter indices. None for no dispersion.

    """
    _scale = None

    def __init__(self, label, center, width, center_dispersion=None,
                 width_dispersion=None, scale=None, normalize=True,
                 backsweep=False, backsweep_period=None):
        self.center = center
        self.center_dispersion = center_dispersion
        self.width = width
        self.width_dispersion = width_dispersion
        self.scale = scale
        self.normalize = normalize
        self.backsweep = backsweep
        self.backsweep_period = backsweep_period
        super(GaussianIrf, self).__init__(label)

    @property
    def scale(self):
        return self._scale

    @scale.setter
    def scale(self, value):
        if not isinstance(value, list):
            raise TypeError("Scale must be list  of parameter indices.")
        if any(not isinstance(val, int) for val in value):
            raise TypeError("Scale must be list  of parameter indices.")
        self._scale = value

    @property
    def center(self):
        return self._center

    @center.setter
    def center(self, value):
        if not isinstance(value, list):
            value = [value]
        self._center = value

    @property
    def center_dispersion(self):
        return self._center_dispersion

    @center_dispersion.setter
    def center_dispersion(self, value):
        if not isinstance(value, list):
            value = [value]
        self._center_dispersion = value

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, value):
        if not isinstance(value, list):
            value = [value]
        self._width = value

    @property
    def width_dispersion(self):
        return self._width_dispersion

    @width_dispersion.setter
    def width_dispersion(self, value):
        if not isinstance(value, list):
            value = [value]
        self._width_dispersion = value

    @property
    def normalize(self):
        return self._normalize

    @normalize.setter
    def normalize(self, value):
        if not isinstance(value, bool):
            raise TypeError("Normalize must be 'true' or 'false'.")
        self._normalize = value

    @property
    def backsweep(self):
        """True or false """
        return self._backsweep

    @backsweep.setter
    def backsweep(self, value):
        """

        Parameters
        ----------
        value : True or False


        Returns
        -------


        """
        if not isinstance(value, bool):
            raise TypeError("Backsweep must be True or False")
        self._backsweep = value

    @property
    def backsweep_period(self):
        """Parameter Index"""
        return self._backsweep_period

    @backsweep_period.setter
    def backsweep_period(self, value):
        """

        Parameters
        ----------
        value : Parameter Index


        Returns
        -------


        """
        self._backsweep_period = value

    def type_string(self):
        t = "'Gaussian'"
        if len(self.center) != len(self.width):
            if len(self.width) is 2:
                t = "'Double Gaussian'"
            elif len(self.width) is 3:
                t = "'Triple Gaussian'"
            elif len(self.width) > 3:
                t = "'{} Gaussian'".format(len(self.width))
        elif len(self.center) is not 1:
            t = "'Multiple Gaussian'"
        return t

    def __str__(self):
        string = super(GaussianIrf, self).__str__()
        string += f"* _Center_: {self.center}\n"
        string += f"* _Width_: {self.width}\n"
        string += f"* _Center Dispersion_: {self.center_dispersion}\n"
        string += f"* _Width Dispersion_: {self.width_dispersion}\n"
        string += f"* _Scale_: {self.scale}\n"
        string += f"* _Nomalize_: {self.normalize}\n"
        return string



import yaml
import os
from .model_spec_yaml import get_model_parser


def parse_file(fname):
    if not os.path.isfile(fname):
        raise Exception("File does not exist.")

    with open(fname) as f:
        spec = load(f)

    return parse_spec(spec)


def parse_yml(spec):
    spec = load(spec)
    return parse_spec(spec)


def parse_spec(spec):

    parser = get_model_parser(spec)

    parser.parse()

    return parser.model


def load(s):
    return yaml.load(s)



"""Glotarans fitmodel package"""

from . import (matrix, matrix_group_generator, fitmodel, result)

FitModel = fitmodel.FitModel
MatrixGroupGenerator = matrix_group_generator.MatrixGroupGenerator
Matrix = matrix.Matrix
Result = result.Result



""" Glotaran Kinetic Matrix """

from typing import List, Tuple
import numpy as np
import lmfit

from glotaran.fitmodel import Matrix
from glotaran.model import Model

from .c_matrix_cython.c_matrix_cython import CMatrixCython
from .irf_gaussian import GaussianIrf
from .irf_measured import MeasuredIrf
from .k_matrix import KMatrix

_BACKEND = CMatrixCython()


class KineticMatrix(Matrix):
    """Implementation of fitmodel.Matrix for a kinetic model."""
    def __init__(self, x: float, dataset: str, model: Model):
        """

        Parameters
        ----------
        x : float
            Point on the estimated axis the matrix calculated for

        dataset : str
            Dataset label of the dataset the matrix is calculated for

        model : glotaran.Model
            The model the matrix is calculated for


        """

        super(KineticMatrix, self).__init__(x, dataset, model)

        self._irf = None
        self._collect_irf(model)
        self._disp_center = dataset.dispersion_center

        self._k_matrices = []
        self._megacomplex_scaling = []
        self._collect_k_matrices(model)

        self._initial_concentration = None
        self._collect_initial_concentration(model)

    @property
    def compartment_order(self) -> List[str]:
        """A list with compartment labels. The index of label indicates the
        index of the compartment in the matrix."""
        compartment_order = [c for mat in self._k_matrices
                             for c in mat.compartment_map]

        compartment_order = list(set(compartment_order))
        return [c for c in self.model.compartments if c in compartment_order]

    @property
    def shape(self) -> Tuple[int, int]:
        """The matrix dimensions as tuple(M, N)."""
        return (self.time.shape[0], len(self.compartment_order))

    def _collect_irf(self, model):
        if self.dataset.irf is None:
            return
        self._irf = model.irfs[self.dataset.irf]

    def _collect_k_matrices(self, model):
        for cmplx in [model.megacomplexes[mc] for mc in self.dataset.megacomplexes]:
            model_k_matrix = None
            for k_matrix_label in cmplx.k_matrices:
                mat = model.k_matrices[k_matrix_label]

                # If multiple k matrices are present, we combine them
                if model_k_matrix is None:
                    model_k_matrix = mat
                else:
                    model_k_matrix = model_k_matrix.combine(mat)
            scaling = self.dataset.megacomplex_scaling[cmplx.label] \
                if cmplx.label in self.dataset.megacomplex_scaling else None
            self._megacomplex_scaling.append(scaling)
            self._k_matrices.append(model_k_matrix)

    def _collect_initial_concentration(self, model):
        if self.dataset.initial_concentration is None:
            return
        self._initial_concentration = \
            model.initial_concentrations[self.dataset.initial_concentration]

    def calculate(self,
                  matrix: np.array,
                  compartment_order: List[str],
                  parameter: lmfit.Parameters):
        """ Calculates the matrix.

        Parameters
        ----------
        matrix : np.array
            The preallocated matrix.

        compartment_order : list(str)
            A list of compartment labels to map compartments to indices in the
            matrix.

        parameter : lmfit.Parameters
            A dictory of parameters.


        Returns
        ^
        -------

        """

        for k_matrix, scale in self._k_matrices_and_scalings():

            scale = parameter.get(scale) if scale is not None else 1.0
            scale *= self._dataset_scaling(parameter)

            self._calculate_for_k_matrix(matrix, compartment_order, k_matrix,
                                         parameter, scale)

    def _k_matrices_and_scalings(self):
        """ Iterator of k matrices and scalings"""
        for i in range(len(self._k_matrices)):
            yield self._k_matrices[i], self._megacomplex_scaling[i]

    def _calculate_for_k_matrix(self,
                                matrix: np.array,
                                compartment_order: List[str],
                                k_matrix: KMatrix,
                                parameter: lmfit.Parameters,
                                scale: str):
        # pylint: disable=too-many-locals
        # pylint: disable=too-many-arguments

        # calculate k_matrix eigenvectos
        eigenvalues, _ = k_matrix.eigen(parameter)
        rates = -eigenvalues

        # we need this since the full c matrix can have more compartments then
        # the k matrix
        compartment_idxs = [compartment_order.index(c) for c in
                            k_matrix.compartment_map]

        # get the time axis
        time = self.dataset.dataset.get_axis("time")

        # calculate the c_matrix
        if isinstance(self._irf, GaussianIrf):
            centers, widths, irf_scale, backsweep, backsweep_period = \
                    self._calculate_irf_parameter(parameter)
            _BACKEND.c_matrix_gaussian_irf(matrix,
                                           compartment_idxs,
                                           rates,
                                           time,
                                           centers, widths,
                                           scale * irf_scale,
                                           backsweep,
                                           backsweep_period,
                                           )

        else:
            _BACKEND.c_matrix(matrix, compartment_idxs, rates, time,
                              scale)
            if isinstance(self._irf, MeasuredIrf):
                irf = self._irf.data
                if len(irf.shape) == 2:
                    idx = (np.abs(self.dataset.data.spectral_axis - self.index)).argmin()
                    irf = irf[idx, :]
                for i in range(matrix.shape[1]):
                    matrix[:, i] = np.convolve(matrix[:, i], irf, mode="same")

        if self._initial_concentration is not None:
            self._apply_initial_concentration_vector(matrix,
                                                     k_matrix,
                                                     parameter,
                                                     compartment_order)

    def _calculate_irf_parameter(self, parameter):

        centers = np.asarray([parameter.get(i) for i in self._irf.center])
        widths = np.asarray([parameter.get(i) for i in self._irf.width])

        center_dispersion = \
            np.asarray([parameter.get(i) for i in self._irf.center_dispersion]) \
            if len(self._irf.center_dispersion) is not 0 else []

        width_dispersion = \
            np.asarray([parameter.get(i) for i in self._irf.width_dispersion]) \
            if len(self._irf.width_dispersion) is not 0 else []

        dist = (self.index - self._disp_center)/100
        if len(center_dispersion) is not 0:
            for i, disp in enumerate(center_dispersion):
                centers = centers + disp * np.power(dist, i+1)

        if len(width_dispersion) is not 0:
            for i, disp in enumerate(width_dispersion):
                widths = widths + disp * np.power(dist, i+1)

        if len(self._irf.scale) is 0:
            scale = np.ones(centers.shape)
        else:
            scale = np.asarray([parameter.get(i) for i in self._irf.scale])

        backsweep = 1 if self._irf.backsweep else 0

        backsweep_period = parameter.get(self._irf.backsweep_period) \
            if self._irf.backsweep else 0

        return centers, widths, scale, backsweep, backsweep_period

    def _apply_initial_concentration_vector(self, c_matrix, k_matrix,
                                            parameter, compartment_order):
        """

        Parameters
        ----------
        c_matrix :

        eigenvectors :

        parameter :

        compartment_order :


        Returns
        -------

        """
        mask = [c in self.compartment_order for c in compartment_order]

        temp = np.dot(np.copy(c_matrix[:, mask]),
                      k_matrix.a_matrix(self._initial_concentration, parameter))

        for i, c in enumerate(self.compartment_order):
            c_matrix[:, compartment_order.index(c)] = temp[:, i]

    @property
    def time(self):
        """The time axis of the matrix """
        return self.dataset.dataset.get_axis("time")

    def _dataset_scaling(self, parameter: lmfit.Parameters):
        """Gets the dataset scaling value from the parameters.

        Parameters
        ----------
        parameter :


        Returns
        -------

        """
        return parameter.get(self.dataset.scaling) if self.dataset.scaling is not None else 1.0



""" Glotaran Kinetic Megacomplex """

from typing import List, Union

from glotaran.model.megacomplex import Megacomplex


class KineticMegacomplex(Megacomplex):
    """A Megacomplex with one or more K-Matrices."""
    def __init__(self, label: str, k_matrices: Union[str, List[str]]):
        """

        Parameters
        ----------
        label: str
            The label of the megacomplex.

        k_matrices: list(str) :
            A list of K-Matrix labels.

        Returns
        -------

        """
        if not isinstance(k_matrices, list):
            k_matrices = [k_matrices]
        self._k_matrices = k_matrices
        super(KineticMegacomplex, self).__init__(label)

    @property
    def k_matrices(self):
        """ K-Matrices associated with the megacomplex."""
        return self._k_matrices

    @k_matrices.setter
    def k_matrices(self, value):
        if not isinstance(value, list):
            value = [value]
        if any(not isinstance(m, str) for m in value):
            raise TypeError
        self._k_matrices = value

    def __str__(self):
        string = super(KineticMegacomplex, self).__str__()
        string += f"* _K-Matrices_: {self.k_matrices}\n"
        return string



"""Spectral Temporal Dataset Descriptor"""

from typing import Dict, List
from glotaran.model import DatasetDescriptor, CompartmentConstraint


class SpectralTemporalDatasetDescriptor(DatasetDescriptor):
    """SpectralTemporalDatasetDescriptor is an implementation of
    model.DatasetDescriptor for spectral or temporal models.

    A SpectralTemporalDatasetDescriptor additionally contains an
    instrument response functions(IRF) and one or more spectral shapes.
    """

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments
    # pylint: disable=attribute-defined-outside-init
    # Datasets are complex.

    def __init__(self,
                 label: str,
                 initial_concentration: str,
                 megacomplexes: List[str],
                 megacomplex_scaling: Dict[str, List[str]],
                 scaling: str,
                 compartment_scaling: Dict[str, List[str]],
                 compartment_constraints: List[CompartmentConstraint],
                 irf: str,
                 shapes: Dict[str, str]):
        """

        Parameters
        ----------
        label : str
            The label of the dataset.

        initial_concentration : str
            The label of the initial concentration

        megacomplexes : List[str]
            A list of megacomplex labels

        megacomplex_scaling : Dict[str: List[str]]
            The megacomplex scaling parameters

        scaling : str
            The scaling parameter for the dataset

        compartment_scaling: Dict[str: List[str]]
            The compartment scaling parameters

        compartment_constraints: List[CompartmentConstraint] :
            A list of compartment constraints

        irf : str
            The label of the dataset's IRF

        shapes : dict(str, str)
            A dictory of comparment and shape label pairs

        """
        self.irf = irf
        self.shapes = shapes
        self.dispersion_center = None
        super(SpectralTemporalDatasetDescriptor, self).\
            __init__(label,
                     initial_concentration,
                     megacomplexes,
                     megacomplex_scaling,
                     scaling,
                     compartment_scaling,
                     compartment_constraints)

    @property
    def irf(self) -> str:
        """The label of the dataset's IRF"""
        return self._irf

    @irf.setter
    def irf(self, irf):
        self._irf = irf

    @property
    def shapes(self) -> Dict[str, str]:
        """A dictory of comparment and shape label pairs"""
        return self._shapes

    @shapes.setter
    def shapes(self, value):
        self._shapes = value

    @property
    def dispersion_center(self):
        """ """
        if self._dispersion_center is None:
            return self.dataset.spectral_axis[0]
        else:
            return self._dispersion_center

    @dispersion_center.setter
    def dispersion_center(self, value):
        """

        Parameters
        ----------
        value :


        Returns
        -------

        """
        self._dispersion_center = value

    def __str__(self):
        string = super(SpectralTemporalDatasetDescriptor, self).__str__()
        if self.irf:
            string += f"* _Irf_: {self.irf}\n"
        if len(self.shapes):
            string += f"* _Shapes_:\n"
            for comp, shape in self.shapes.items():
                string += f"  * __{comp}__: {shape}\n"
        return string



"""Glotaran Fitmodel Result"""

import numpy as np
from lmfit_varpro import SeparableModelResult

from glotaran.model.dataset import Dataset
from glotaran.model.parameter_group import ParameterGroup


class Result(SeparableModelResult):
    """The result of fit."""

    @property
    def best_fit_parameter(self) -> ParameterGroup:
        """The best fit parameters."""
        return ParameterGroup.from_parameter_dict(self._result.params)

    def estimated_matrix(self, dataset: str) -> np.array:
        """Returns the estimated matrix of the model.

        Parameters
        ----------
        dataset: str
            Label of the dataset.


        Returns
        -------
        e_matrix: np.array
            Estimated Matrix

        """
        return np.asarray(self.e_matrix(**{"dataset": dataset}))

    def calculated_matrix(self, dataset: str) -> np.array:
        """Returns the calculated matrix of the model.

        Parameters
        ----------
        dataset: str
            Label of the dataset.


        Returns
        -------
        c_matrix: np.array
            Calculated Matrix

        """
        return np.asarray(self.c_matrix(**{"dataset": dataset}))

    def fitted_data(self, dataset: str) -> Dataset:
        """Returns the fitted dataset.

        Parameters
        ----------
        dataset: str
            Label of the dataset.


        Returns
        -------
        fitted_dataset: glotaran.Dataset

        """
        data = np.asarray(self.eval(**{"dataset": dataset}))
        dataset = self.model.datasets[dataset].dataset.copy()
        dataset.set(data)
        return dataset

    @property
    def model(self) -> 'glotaran.Model':
        """The Glotaran Model used to fit the data."""
        return self.get_model().model

    def __str__(self):
        string = "# Fitresult\n\n"

        # pylint: disable=invalid-name

        ll = 32
        lr = 13

        string += "Optimization Result".ljust(ll-1)
        string += "|"
        string += "|".rjust(lr)
        string += "\n"
        string += "|".rjust(ll, "-")
        string += "|".rjust(lr, "-")
        string += "\n"

        string += "Number of residual evaluation |".rjust(ll)
        string += f"{self._result.nfev} |".rjust(lr)
        string += "\n"
        string += "Number of variables |".rjust(ll)
        string += f"{self._result.nvarys} |".rjust(lr)
        string += "\n"
        string += "Number of datapoints |".rjust(ll)
        string += f"{self._result.ndata} |".rjust(lr)
        string += "\n"
        string += "Negrees of freedom |".rjust(ll)
        string += f"{self._result.nfree} |".rjust(lr)
        string += "\n"
        string += "Chi Square |".rjust(ll)
        string += f"{self._result.chisqr:.6f} |".rjust(lr)
        string += "\n"
        string += "Reduced Chi Square |".rjust(ll)
        string += f"{self._result.redchi:.6f} |".rjust(lr)
        string += "\n"

        string += "\n"
        string += "## Best Fit Parameter\n\n"
        string += f"{self.best_fit_parameter}"
        string += "\n"

        return string



from enum import Enum


class ParameterConstraintType(Enum):
    fix = 0
    bound = 1


class ParameterConstraint(object):
    """
    A parameter constraint concerns one more parameters and has a type.
    """
    def __init__(self, parameter):
        self._range = None
        if isinstance(parameter, tuple):
            self.range = parameter
        else:
            self.parameter = parameter

    @property
    def range(self):
        return self._range

    @range.setter
    def range(self, val):
        if not isinstance(val, tuple):
            raise TypeError("Range must be tuple")
        if not len(val) == 2:
            raise ValueError("Range must be (FROM, TO)")
        self.parameter = list(range(val[0], val[1]))
        self._range = val

    @property
    def parameter(self):
        return self._parameter

    @parameter.setter
    def parameter(self, value):
        if not isinstance(value, list):
            value = [value]
        if any(not isinstance(v, int) for v in value):
            raise TypeError("Parameter mus be an integer or a list of integer")
        self._parameter = value

    def type(self):
        raise NotImplementedError

    def type_string(self):
        raise NotImplementedError

    def __str__(self):
        if self.range is None:
            parameter = self.parameter
        else:
            parameter = self.range
        return "Type: '{}' Parameters: {}".format(self.type_string(),
                                                  parameter)


class FixedConstraint(ParameterConstraint):
    def type(self):
        return ParameterConstraintType.fix

    def type_string(self):
        return "Fixed"


class BoundConstraint(ParameterConstraint):
    """
    Represents a boundary constraint.
    """
    def __init__(self, parameters, lower='NaN', upper='NaN'):
        self.lower = lower
        self.upper = upper
        super(BoundConstraint, self).__init__(parameters)

    @property
    def lower(self):
        return self._lower

    @lower.setter
    def lower(self, value):
        if not isinstance(value, (int, float)) and value != 'NaN':
            raise TypeError("Lower bound must be a number oder 'NaN'.")
        self._lower = value

    @property
    def upper(self):
        return self._upper

    @upper.setter
    def upper(self, value):
        if not isinstance(value, (int, float)) and value != 'NaN':
            raise TypeError("Upper bound must be a number or 'NaN'.")
        self._upper = value

    def type(self):
        return ParameterConstraintType.bound

    def type_string(self):
        return "Bound"

    def __str__(self):
        return "{} Lower: {} Upper: {}".format(super(BoundConstraint,
                                                     self).__str__(),
                                               self.lower, self.upper)



"""Glotaran Fitmodel Matrix"""

from typing import List, Tuple
from abc import ABC, abstractmethod
import numpy as np

from glotaran.model.compartment_constraints import CompartmentConstraintType
from glotaran.model.dataset_descriptor import DatasetDescriptor
from glotaran.model.parameter_group import ParameterGroup


class Matrix(ABC):
    """An abstract matrix which gets prepared with a dataset and a model and
    can be calculated by giving a ParameterGroup."""

    def __init__(self, index: any, dataset: DatasetDescriptor, model:
                 'glotaran.Model'):
        """

        Parameters
        ----------
        x: any
            Index in the matrix group.

        dataset : glotaran.DatasetDescriptor
            Dataset descriptor for the matrix.

        model: glotaran.Model
            Model for the matrix.


        """
        self.index = index
        self._dataset = dataset
        self._model = model

    def calculate_standalone(self, parameter: ParameterGroup) -> np.array:
        """ Calculates the matrix outside a matrix group.

        Parameters
        ----------
        parameter : glotaran.ParameterGroup
            Parameters to calculate the matrix with.


        Returns
        -------
        matrix : numpy.array
            The calculated matrix.

        """
        matrix = np.zeros((self.shape), np.float64)
        self.calculate(matrix, self.compartment_order, parameter)
        self.apply_constraints(matrix, self.compartment_order, parameter)
        return matrix

    def apply_constraints(self,
                          matrix: np.ndarray,
                          compartment_order: List[str],
                          parameter: ParameterGroup):
        """Applies Zero and Equal constraints to the matrix.

        Parameters
        ----------
        matrix: np.array
            Target matrix.

        compartment_order: List[str]
            A list of compartments representing the  mapping of compartments
            to entries in the target matrix.

        parameter: ParameterGroup
            Parameters to calculate the matrix with.

        """
        # get constraint compartment indeces
        constraints = [c for c in self.dataset.compartment_constraints
                       if c.applies(self.index)
                       and c.type() is not CompartmentConstraintType.equal_area]
        for constrain in constraints:
            idx = compartment_order.index(constrain.compartment)

            matrix[:, idx].fill(0.0)
            if constrain.type() == CompartmentConstraintType.equal:
                for target, param in constrain.targets_and_parameter:
                    t_idx = compartment_order.index(target)
                    param = parameter.get(param)
                    matrix[:, idx] += param * matrix[:, t_idx]

    @property
    def dataset(self) -> DatasetDescriptor:
        """The underlying glotaran.DatasetDescriptor"""
        return self._dataset

    @property
    def model(self) -> 'glotaran.Model':
        """The underlying glotaran.Model"""
        return self._model

    @property
    def compartment_order(self) -> List[str]:
        """A list of compartments representing the internal mapping of
        compartments to entries in the matrix."""
        raise NotImplementedError

    @property
    def shape(self) -> Tuple[int, int]:
        """A tuple representing the dimensions. First entry is the compartment
        index. Use Matrix.compartment_order to map entries to compartments."""
        raise NotImplementedError

    @abstractmethod
    def calculate(self, matrix: np.array, compartment_order: List[str], parameter: ParameterGroup):
        """Calculates the matrix.

        Parameters
        ----------
        matrix: np.array
            Target matrix.

        compartment_order: List[str]
            A list of compartments representing the  mapping of compartments
            to entries in the target matrix.

        parameter: ParameterGroup
            Parameters to calculate the matrix with.

        """
        raise NotImplementedError



from enum import Enum
from cycler import cycler
import matplotlib.colors as colors


def hex_to_rgb(hex_string):
    rgb = colors.hex2color(hex_string)
    return tuple([int(255*x) for x in rgb])


def rgb_to_hex(rgb_tuple):
    return colors.rgb2hex([1.0*x/255 for x in rgb_tuple])


class GlotaranDefaultColorCode(Enum):
    # Name	#Hex
    black = '#000000'
    red = '#ff0000'
    blue = '#0000ff'
    green = '#00ff00'
    magenta = '#ff00ff'
    cyan = '#00ffff'
    yellow = '#ffff00'
    green4 = '#008b00'
    orange = '#ff8c00'
    brown = '#964b00'
    grey = '#808080'
    violet = '#9400d3'
    turquoise = '#40e0d0'
    maroon = '#800000'
    indigo = '#4b0082'


class GlotaranLineStyles(Enum):
    solid = '-'
    dashed = '--'
    dotted = '..'
    dashdot = '-.'


def get_glotaran_default_colors():
    return [e.value for e in GlotaranDefaultColorCode]


def get_glotaran_default_colors_cycler():
    return cycler('color', [e.value for e in GlotaranDefaultColorCode])



"""Glotarans dataIO package"""

from . import wavelength_time_explicit_file

WavelengthExplicitFile = wavelength_time_explicit_file.WavelengthExplicitFile
TimeExplicitFile = wavelength_time_explicit_file.TimeExplicitFile



from glotaran.models.damped_oscillation import (
    DOASModel,
    DOASMegacomplex,
    Oscillation)

from .model_spec_yaml import (Keys,
                              register_model_parser)
from .model_spec_yaml_kinetic import KineticKeys, KineticModelParser
from .utils import get_keys_from_object


class DOASKeys:
    OSCILLATIONS = "oscillations"
    FREQUENCY = "frequency"
    RATE = "rate"


class DOASModelParser(KineticModelParser):
    def get_model(self):
        return DOASModel

    def get_additionals(self):
        super(DOASModelParser, self).get_additionals()
        self.get_oscillations()

    def get_megacomplexes(self):
        for cmplx in self.spec[Keys.MEGACOMPLEXES]:
            (label, mat, osc) = get_keys_from_object(cmplx,
                                                     [Keys.LABEL,
                                                      KineticKeys.K_MATRICES,
                                                      DOASKeys.OSCILLATIONS,
                                                      ])
            mat = [] if mat is None else mat
            osc = [] if osc is None else osc
            self.model.add_megacomplex(DOASMegacomplex(label, mat, osc))

    def get_oscillations(self):
        if DOASKeys.OSCILLATIONS not in self.spec:
            return
        for osc in self.spec[DOASKeys.OSCILLATIONS]:
            (label, comp, freq, rate) = get_keys_from_object(osc,
                                                             [Keys.LABEL,
                                                              Keys.COMPARTMENT,
                                                              DOASKeys.FREQUENCY,
                                                              DOASKeys.RATE,
                                                              ])
            self.model.add_oscillation(Oscillation(label, comp, freq, rate))


register_model_parser("dampened_oscillation", DOASModelParser)



from glotaran.model import Dataset
# from .spectral_timetrace import SpectralTimetrace, SpectralUnit
from enum import Enum
import os.path
import numpy as np
import pandas as pd


class DataFileType(Enum):
    mlsd_mulheim = "MLSD Mulheim"


class HeaderMLSDMulheim(Enum):
    int_time = "[integration time:]"
    int_delay = "[integration delay:]"
    boxcar = "[boxcar width:]"
    avg = "[average:]"
    dark = "[dark spectrum:]"
    comment = "[comment:]"
    prot = "[number, period, cycle, actinic light]"
    data = "[data, wavelength(once), 1st dark, 2nd dark, measure]"


class MLSDFile(object):
    """
    Class capable of reading in so called time- or wavelength-explicit file format.
    Returns a glotaran.model.dataset
    """
    def __init__(self, file, debug=False):
        self._file = file
        self._file_data_format = None
        self._observations = []  # TODO: choose name: data_points, observations, data
        self._times = []
        self._spectral_indices = []
        self._label = ""
        self._comment = ""
        self._debug = debug

    def get_explicit_axis(self):
        raise NotImplementedError

    def set_explicit_axis(self, axis):
        raise NotImplementedError

    def get_secondary_axis(self):
        raise NotImplementedError

    def get_data_row(self, index):
        raise NotImplementedError

    def get_observations(self, index):
        raise NotImplementedError

    def get_format_name(self):
        raise NotImplementedError

    def write(self, filename, overwrite=False, comment="", file_format="Time explicit"):
        # TODO: write a more elegant method

        if os.path.isfile(filename) and not overwrite:
            print('File {} already exists'.format(os.path.isfile(filename)))
            raise Exception("File already exist.")

        comments = "# Filename: " + self._file + "\n" + " ".join(self._comment.splitlines()) + "\n"

        if file_format == "Wavelength explicit":
            wav = '\t'.join([repr(num) for num in self._spectral_indices])
            header = comments + "Wavelength explicit\nIntervalnr {}" \
                                "".format(len(self._spectral_indices)) + "\n" + wav
            raw_data = np.vstack((self._times.T, self._observations)).T
        elif file_format == "Time explicit":
            tim = '\t'.join([repr(num) for num in self._times])
            header = comments + "Time explicit\nIntervalnr {}" \
                                "".format(len(self._times)) + "\n" + tim
            raw_data = np.vstack((self._spectral_indices.T, self._observations.T)).T
        else:
            raise NotImplementedError

        np.savetxt(filename, raw_data, fmt='%.18e', delimiter='\t', newline='\n',
                   header=header, footer='', comments='')

    def read(self, label):
        if not os.path.isfile(self._file):
            raise Exception("File does not exist.")

        self._label = label
        with open(self._file) as f:
            self._file_data_format = get_data_file_format(f)  # TODO: what to do with return: None?
            if not self._file_data_format:
                raise ImportError

            self.read_comment(f)
            self.read_prot(f)
            self.read_data(f)

    def dataset(self):
        dataset = Dataset(self._label)
        if not self._file_data_format:
            return None
        elif self._file_data_format == DataFileType.mlsd_mulheim:
            dataset.set_axis("spec", self._spectral_indices)
            dataset.set_axis("time", self._times)
        dataset.data = self._observations
        return dataset

    def read_comment(self, f):
        f.seek(0)
        comment = []
        for line in f:
            if line.startswith("[") and line.startswith(str(HeaderMLSDMulheim.comment)):
                break
        for line in f:
            if line.startswith("["):
                break
            elif not line.strip():
                pass
            else:
                comment.append(line)
        self._comment = self._file + '\n' + '_'.join(comment)

    def read_prot(self, f):
        f.seek(0)
        protocol = []
        for line in f:
            if line.startswith("[") and line.startswith(HeaderMLSDMulheim.prot.value):
                break
        for line in f:
            if line.startswith("["):
                break
            elif not line.strip():
                pass
            else:
                protocol.append(line)
        tmp = [0]
        for l in protocol:
            val = l.strip().split()
            period = float(val[1])
            cycle = float(val[2])
            for i in range(int(cycle)):
                tmp.append(tmp[-1]+period)
            times = np.array(tmp)
        if self._debug:
            print('len(times)={}'.format(len(times)))

        self._times = times

    def read_data(self, f):
        f.seek(0)
        line = f.readline()
        while line:
            if line.startswith("[") and line.startswith(HeaderMLSDMulheim.data.value):
                break
            line = f.readline()
        df = pd.read_table(f, header=None, index_col=None)
        raw_data = df.values
        self._spectral_indices = raw_data[:, 0]
        data = np.empty((raw_data.shape[0], int((raw_data.shape[1]-1)/3)))
        for i, j in zip(range(0, data.shape[1], 1), range(1, raw_data.shape[1], 3)):
            data[:, i] = raw_data[:, j+2] - (raw_data[:, j]+raw_data[:, j+1])/2
        self._observations = data
        # self._times = range(0,data.shape[1])


def get_data_file_format(f):
    required_headers = ("[number, period, cycle, actinic light]",
                        "[data, wavelength(once), 1st dark, 2nd dark, measure]")
    nmatches = 0
    data_file_format = None
    line = ""
    while not data_file_format:
        line = f.readline()
        if line.startswith("["):
            if any(substring in line for substring in required_headers):
                nmatches += 1
                if nmatches > 1:
                    data_file_format = DataFileType.mlsd_mulheim
                    break

    return data_file_format


# def is_blank (mystr):
#     return not (mystr and mystr.strip())


# def is_not_blank (mystr):
#     return bool(mystr and mystr.strip())



import numpy as np

from glotaran.fitmodel import Matrix

from .spectral_shape_gaussian import SpectralShapeGaussian


class SpectralMatrix(Matrix):
    def __init__(self, x, dataset, model):
        super(SpectralMatrix, self).__init__(x, dataset, model)

        self.shapes = {}
        self.collect_shapes()

    def collect_shapes(self):

        for c, shape in self.dataset.shapes.items():
            self.shapes[c] = self.model.shapes[shape]

    @property
    def compartment_order(self):
        cmplxs = [self.model.megacomplexes[c] for c in self.dataset.megacomplexes]
        kmats = [self.model.k_matrices[k] for cmplx in cmplxs
                 for k in cmplx.k_matrices]
        return list(set([c for kmat in kmats for c in kmat.compartment_map]))

    @property
    def shape(self):
        x = self.dataset.dataset.spectral_axis
        return (x.shape[0], len(self.compartment_order))

    def calculate(self, c_matrix, compartment_order, parameter):

        # We need the spectral shapes and axis to perform the calculations
        x = self.dataset.dataset.spectral_axis

        for (i, c) in enumerate(compartment_order):
            if c in self.shapes:
                c_matrix[:, i] = self._calculate_shape(parameter, self.shapes[c], x)
            else:
                # we use ones, so that if no shape is defined for the
                # compartment, the  amplitude is 1.0 by convention.
                c_matrix[:, i].fill(1.0)

    def _calculate_shape(self, parameter, shape, x):
        if isinstance(shape, SpectralShapeGaussian):
            amp = parameter.get(shape.amplitude)
            location = parameter.get(shape.location)
            width = parameter.get(shape.width)
            return amp * np.exp(-np.log(2) *
                                np.square(2 * (x - location)/width))
        raise ValueError(f"uknown shape '{type(shape)}'")



import cython
cimport cython

import numpy as np
cimport numpy as np

from libc.math cimport exp
from numpy.math cimport NAN

def __init__():
    np.import_array()


@cython.boundscheck(False)
@cython.wraparound(False)
def calculateC(double[:, :] C, idxs, double[:] k, double[:] T, double scale):
    nr_times = T.shape[0]
    nr_comps = k.shape[0]
    cdef int n_c, n_t, n_k
    cdef double t_n, k_n
    for n_k in range(nr_comps):
        n_c = idxs[n_k]
        k_n = k[n_k]
        for n_t in range(nr_times):
            t_n = T[n_t]
            C[n_t, n_c] += scale * exp(-k_n * t_n)



""" Glotaran Fitmodel Matrix Group Generator"""

from typing import List, Generator, Tuple, Type
from collections import OrderedDict
import numpy as np

from glotaran.model.dataset_descriptor import DatasetDescriptor
from glotaran.model.parameter_group import ParameterGroup

from .matrix import Matrix
from .matrix_group import MatrixGroup


class MatrixGroupGenerator:
    """A MatrixGroupGenerator groups a set of matrices along the estimated or
    calculated axis.
    """
    def __init__(self, matrix: Type[Matrix], calculated=False):
        """

        Parameters
        ----------
        matrix : Type[Matrix]
            An implementation of fitmodel.Matrix
        calculated :
            (Default value = False)
            If true, group along the calculated insted of eastimated axis

        """
        self._groups = OrderedDict()
        self._matrix = matrix
        self._calculated = calculated

    @classmethod
    def for_model(cls, model: 'glotaran.Model', matrix: Type[Matrix], xtol=0.5, calculated=False):
        """ Creates matrix group generator for a full model.

        Parameters
        ----------
        model : glotaran.Model
            A Glotaran Model
        matrix : Type[fitmodel.Matrix]
            An implementation of fitmodel.Matrix
        xtol :
            (Default value = 0.5)
            Grouping tolerance
        calculated :
            (Default value = False)
            If true, group along the calculated insted of eastimated axis

        """
        gen = cls(matrix, calculated)
        gen.init_groups_for_model(model, xtol)
        return gen

    @classmethod
    def for_dataset(cls,
                    model: 'glotaran.Model',
                    dataset: str,
                    matrix: Type[Matrix],
                    calculated=False):
        """ Creates matrix group generator for a dataset.

        Parameters
        ----------
        model : glotaran.Model
            A Glotaran Model

        dataset: str
            Label of the dataset

        matrix : Type[fitmodel.Matrix]
            An implementation of fitmodel.Matrix

        calculated :
             (Default value = False)
            If true, group along the calculated insted of eastimated axis

        """
        gen = cls(matrix, calculated)
        data = model.datasets[dataset]
        gen.add_dataset_to_group(model, data, 0)
        return gen

    def init_groups_for_model(self, model: 'glotaran.Model', xtol: float):
        """ Initializes the matrix groups for a model.

        Parameters
        ----------
        model : glotaran.Model
            A Glotaran Model

        xtol :
            Grouping tolerance

        """
        for _, dataset in model.datasets.items():
            self.add_dataset_to_group(model, dataset, xtol)

    def add_dataset_to_group(self,
                             model: 'glotaran.Model',
                             dataset: DatasetDescriptor,
                             xtol: float):
        """ Adds a dataset to the group

        Parameters
        ----------
        model : glotaran.Model
            A Glotaran Model

        dataset: glotaran.DatasetDescriptor
            Dataset descriptor

        xtol :
            Grouping tolerance

        """
        grouping_axis = dataset.dataset.get_calculated_axis() if self._calculated\
            else dataset.dataset.get_estimated_axis()
        for matrix in [self._matrix(x, dataset, model) for x
                       in grouping_axis]:
            self._add_matrix_to_group(matrix, xtol)

    def _add_matrix_to_group(self, matrix: Matrix, xtol: float):
        if matrix.index in self._groups:
            self._groups[matrix.index].add_matrix(matrix)
        elif any(abs(matrix.index-val) < xtol for val in self._groups):
            idx = [val for val in self._groups if abs(matrix.index-val) <
                   xtol][0]
            self._groups[idx].add_matrix(matrix)
        else:
            self._groups[matrix.index] = MatrixGroup(matrix)

    def groups(self) -> Generator[MatrixGroup, None, None]:
        """Generator returning all groups."""
        for _, group in self._groups.items():
            yield group

    def groups_in_range(self, rng: Tuple[float, float]) -> List[MatrixGroup]:
        """Returns a list of all groups within a range.

        Parameters
        ----------
        rng : tuple(float, float)
            Range


        Returns
        -------
        groups : list(fitmodel.MatrixGroup)
            All groups in range.


        """
        return [g for g in self.groups() if rng[0] <= g.x <= rng[1]]

    def calculate(self, parameter: ParameterGroup) -> List[np.array]:
        """ Calculates the matrices in the group.

        Parameters
        ----------
        parameter : glotaran.ParameterGroup

        Returns
        -------
        matrices : list(numpy.array)

        """
        return [group.calculate(parameter) for group in self.groups()]

    def create_dataset_group(self) -> List[np.array]:
        """ Groups all datasets in the model and concats them on the calculated
        axis

        Returns
        -------
        datasets : list(numpy.array)
            A list with the grouped datasets
        """

        dataset_group = []
        for _, group in self._groups.items():
            data = np.array([])
            for mat in group.matrices:
                index = np.where(mat.dataset.dataset.get_estimated_axis() ==
                                 mat.index)
                data = np.concatenate((data,
                                       mat.dataset.dataset.get()[index, :].flatten()))
            dataset_group.append(data)
        return dataset_group



import matplotlib
import matplotlib.pyplot as plt
from numpy import linalg
from .glotaran_color_codes import get_glotaran_default_colors_cycler, get_glotaran_default_colors
from cycler import cycler

from glotaran.models.spectral_temporal import KineticModel

# TODO: calculate svd when plots are requested
# TODO: calculate svd in background


def plot_data(*args, **kwargs):
    if len(args) == 1 and isinstance(args[0], KineticModel):
        _plot_data_from_kin_sep_model(args[0])
    elif len(args) == 2 and isinstance(args[1], int):
        pass
    elif len(args) == 4 and isinstance(args[0], matplotlib.axes.Axes):
        _plot_data(args[0], args[1], args[2], args[3])


def _plot_data_from_kin_sep_model(model):
    times = model.datasets['dataset1'].data.get_axis("time")
    spectral_indices = model.datasets['dataset1'].data.get_axis("spec")
    data = model.datasets['dataset1'].data.data.T
    plt.pcolormesh(times, spectral_indices, data)


def _plot_data(ax, times, spectral_indices, data):
    ax.pcolormesh(times, spectral_indices, data)


def plot_data_overview(times, spectral_indices, data):
    fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(nrows=2, ncols=2)
    ax2.set_prop_cycle(get_glotaran_default_colors_cycler())
    ax3.set_prop_cycle(get_glotaran_default_colors_cycler())
    ax4.set_prop_cycle(cycler('color', get_glotaran_default_colors()))

    xmin = min(times)
    xmax = max(times)
    ymin = min(spectral_indices)
    ymax = max(spectral_indices)

    U1, s1, V1 = linalg.svd(data)
    print("U.shape: {} ; V.shape: {} ; s.shape: {}".format(U1.shape, V1.shape, s1.shape))

    ax1.set_xlim([xmin, xmax])
    ax1.set_ylim([ymin, ymax])
    ax1.set_xlabel('Times (ps)')
    ax1.set_ylabel('$Wavenumber\ [\ cm^{-1}\ ]$')

    plt.tight_layout()
    plot_data(ax1, times, spectral_indices, data)

    ax3.set_xlim([xmin, xmax])
    # ax2.set_ylim([ymin, ymax])
    ax3.set_xlabel('Times (ps)')
    ax3.set_ylabel('$Wavenumber\ [\ cm^{-1}\ ]$')
    plot_trace(ax3, times, V1[0, :].T)

    # ax2.set_ylim([xmin, xmax])
    # ax2.set_ylim([ymin, ymax])
    # ax2.set_xlabel('Times (ps)')
    # ax2.set_ylabel('$Wavenumber\ [\ cm^{-1}\ ]$')
    plot_trace(ax2, U1[:, 0:3], spectral_indices)

    plot_trace(ax4, range(3), s1[0:3])

    # ax1.pcolormesh(times, spectral_indices, data)
    # ax1.set_xlim([xmin, xmax])
    # ax1.set_ylim([ymin, ymax])
    # plot_sing_val_svd

    plt.show(block=False)


def plot_trace(ax, x_values, y_values):
    ax.plot(x_values, y_values)


def plot_residuals():
    pass


def plot_residuals_svd():
    pass


def plot_results():
    pass



from enum import Enum


class CompartmentConstraintType(Enum):
    """ """
    zero = 0
    equal = 1
    equal_area = 2


class CompartmentConstraint(object):
    """A compartment constraint is applied on one compartment on one or many
    intervals on the estimated axies type.

    Parameters
    ----------
    compartment: label of the compartment
    intervals: list of tuples representing intervals on the estimated axies
    """
    def __init__(self, compartment, intervals):
        self.compartment = compartment
        self.intervals = intervals

    @property
    def compartement(self):
        """label of the compartment"""
        return self._compartement

    @compartement.setter
    def compartment(self, value):
        """

        Parameters
        ----------
        value : label of the compartment

        """
        self._compartement = value

    @property
    def intervals(self):
        """ """
        return self._intervals

    @intervals.setter
    def intervals(self, value):
        """

        Parameters
        ----------
        value : list of tuples representing intervals on the estimated axies
        """
        if not isinstance(value, list):
            value = [value]
        if any(not isinstance(val, tuple) for val in value):
            raise TypeError("Intervals must be tuples or list of tuples.")
        if any(len(val) is not 2 for val in value):
            raise ValueError("Intervals must be of length 2")
        self._intervals = value

    def applies(self, x):
        """
        Returns true if x is in one of the intervals.
        Parameters
        ----------
        x : point on the estimated axies

        """
        return any(interval[0] <= x <= interval[1] for interval in self.intervals)

    def type(self):
        """ """
        raise NotImplementedError

    def type_string(self):
        """ """
        raise NotImplementedError

    def __str__(self):
        return "Compartment: {}\tType: '{}'\tIntervals: {}"\
                .format(self.compartment,
                        self.type_string(),
                        self.intervals)


class ZeroConstraint(CompartmentConstraint):
    """A zero constraint sets the c matrix of a compartment to 0.

    Parameters
    ----------
    compartment: label of the compartment
    intervals: list of tuples representing intervals on the estimated axies
    """
    def type(self):
        """ """
        return CompartmentConstraintType.zero

    def type_string(self):
        """ """
        return "Zero"


class EqualConstraint(CompartmentConstraint):
    """An equal constraint The compartments c matrix will be replaced with a sum
    of target compartments, each scaled by a parameter.

    C = p1 * C_t1 + p2 * C_t1 + ...

    Parameters
    ----------
    compartment: label of the compartment
    intervals: list of tuples representing intervals on the estimated axies
    targets: list of target compartments
    parameters: list of scaling parameter for the targets
    """
    def __init__(self, compartment, intervals, targets, parameters):
        self.targets = targets
        self.parameters = parameters
        super(EqualConstraint, self).__init__(compartment, intervals)

    def type(self):
        """ """
        CompartmentConstraintType.equal

    @property
    def targets(self):
        """list of target compartments"""
        return self._targets

    @targets.setter
    def targets(self, value):
        """

        Parameters
        ----------
        value :list of target compartments
        """
        if not isinstance(value, list):
            value = [value]
        self._targets = value

    @property
    def parameters(self):
        """list of scaling parameter for the targets"""
        return self._parameters

    @parameters.setter
    def parameters(self, value):
        """

        Parameters
        ----------
        value : list of scaling parameter for the targets
        """
        if not isinstance(value, list):
            value = [value]
        if not len(value) == len(self.targets):
            raise ValueError("number of parameters({}) != nr"
                             " targets({})".format(len(value),
                                                   len(self.targets)))
        self._parameters = value

    def parameter_and_targets(self):
        """generates traget and parameter pairs """
        for i in range(len(self.parameters)):
            yield self.parameters[i], self.targets[i]

    def type_string(self):
        """ """
        return "Equal"

    def __str__(self):
        return "{} Target: {} Parameter: {}".format(super(EqualConstraint,
                                                          self).__str__(),
                                                    self.target,
                                                    self.parameter)


class EqualAreaConstraint(CompartmentConstraint):
    """An equal area constraint adds a the differenc of the sum of a
    compartements in the e matrix in one ore more intervals to the scaled sum
    of the e matrix of one or more target compartmants to resiudal. The additional
    residual is scaled with the weight.

    Parameters
    ----------
    compartment: label of the compartment
    intervals: list of tuples representing intervals on the estimated axies
    targets: list of target compartments
    parameters: list of scaling parameter for the targets
    weight: scaling factor for the residual
    """
    def __init__(self, compartment, intervals, target, parameter, weight):
        self.parameter = parameter
        self.target = target
        self.weight = weight
        super(EqualAreaConstraint, self).__init__(compartment, intervals)

    @property
    def weight(self):
        """scaling factor for the residual"""
        return self._weight

    @weight.setter
    def weight(self, value):
        """

        Parameters
        ----------
        value : scaling factor for the residual
        """
        if not isinstance(value, float):
            raise TypeError("Weight must be float.")
        self._weight = value

    @property
    def target(self):
        """target compartment"""
        return self._target

    @target.setter
    def target(self, value):
        """

        Parameters
        ----------
        value : target compartment
        """
        self._target = value

    @property
    def parameter(self):
        """scaling parameter for the target"""
        return self._parameter

    @parameter.setter
    def parameter(self, value):
        """

        Parameters
        ----------
        value : scaling parameter for the target
        """
        self._parameter = value

    def type(self):
        """ """
        return CompartmentConstraintType.equal_area

    def type_string(self):
        """ """
        return "Equal Area"

    def __str__(self):
        return "{}\tTarget: {}\tParameter: {}\tWeight: {}"\
                .format(super(EqualAreaConstraint, self).__str__(),
                        self.target, self.parameter, self.weight)



from glotaran.model import Dataset


class SpectralTemporalDataset(Dataset):
    """Implementation of model.Dataset for spectral and temporal models.
    Provides a time and a spectral axis.
    """

    def __init__(self, label, time_unit="s", spectral_unit="nm"):
        self.time_unit = time_unit
        self.spectral_unit = spectral_unit
        super(SpectralTemporalDataset, self).__init__(label)

    _time_units = {
        "h":   3600,
        "m":     60,
        "s":      1,
        "ms":  1e-3,
        "us":  1e-6,
        "ns":  1e-9,
        "fs": 1e-15,
    }

    supported_time_units = [u for u in _time_units]
    """supported timeunits """

    _spectral_units = {
        "um":     1e3,
        "nm":       1,
        "cm^-1": 10e6,
    }

    supported_spectral_units = [u for u in _spectral_units]
    "supported spectral units"

    @property
    def time_unit(self):
        """the time unit [default 's'] """
        return self._time_unit

    @time_unit.setter
    def time_unit(self, value):
        if value not in self._time_units:
            raise ValueError("Unknown time unit {}. Supported units are "
                             "{}".format(value, self.supported_time_units))
        self._time_unit = value

    def get_time_axis(self, unit=None):
        """Returns the time axis, does unit conversion if necessary

        Parameters
        ----------
        unit : the desired time unit
             (Default value = None)

        Returns
        -------
        The (converted) time axis
        """
        if unit is None or unit == self.time_unit:
            return self.get_axis("time")
        conversion = self._time_units[self.time_unit] / self._time_units[unit]
        return conversion * self.get_axis("time")

    @property
    def time_axis(self, unit=None):
            return self.get_axis("time")

    @time_axis.setter
    def time_axis(self, value):
        self.set_axis("time", value)

    @property
    def spectral_unit(self):
        """the spectral unit [default: 'nm'] """
        return self._spectral_unit

    @spectral_unit.setter
    def spectral_unit(self, value):
        if value not in self._spectral_units:
            raise ValueError("Unknown spectral unit {}. Supported units are "
                             "{}".format(value, self.supported_spectral_units))
        self._spectral_unit = value

    def get_spectral_axis(self, unit=None):
        """spec spectral axis

        Parameters
        ----------
        unit :
             (Default value = None)

        Returns
        -------

        """
        if unit is None or unit == self.spectral_unit:
            return self.get_axis("spectral")

        if self.spectral_unit in ["nm", "um"]:
            if unit == "cm^-1":
                conversion = self._spectral_units["cm^-1"] * \
                    self._spectral_units[self.spectral_unit]
                return [conversion/v for v in self.get_axis("spectral")]
            conversion = self._spectral_units[self.spectral_unit] / \
                self._spectral_units(unit)
            return conversion * self.get_axis("spectral")
        conversion = self._spectral_units["cm^-1"] / \
            self._spectral_units[self.spectral_unit]
        return [conversion/v for v in self.get_axis("spectral")]

    @property
    def spectral_axis(self, unit=None):
        """the spectral axies """
        return self.get_axis("spectral")

    @spectral_axis.setter
    def spectral_axis(self, value):
        self.set_axis("spectral", value)

    def get_estimated_axis(self):
        """ """
        return self.spectral_axis

    def get_calculated_axis(self):
        """ """
        return self.time_axis



"""Glotaran Model"""


from abc import ABC, abstractmethod
from typing import List, Type, Dict, Generator
from collections import OrderedDict
import numpy as np

from .dataset import Dataset
from .dataset_descriptor import DatasetDescriptor
from .initial_concentration import InitialConcentration
from .megacomplex import Megacomplex
from .parameter_group import ParameterGroup

from glotaran.fitmodel import FitModel, Matrix, Result

ROOT_BLOCK_LABEL = "p"


class Model(ABC):
    """Model represents a global analysis model."""

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments
    # pylint: disable=attribute-defined-outside-init
    # Models are complex.
    def __init__(self):
        """ """
        self._compartments = None
        self._datasets = OrderedDict()
        self._initial_concentrations = None
        self._megacomplexes = {}
        self._parameter = None

    @abstractmethod
    def type_string(self) -> str:
        """Returns a human readable string identifying the type of the model.

        Returns
        -------

        type : str
            Type of the Model

        """
        raise NotImplementedError

    @abstractmethod
    def calculated_matrix(self) -> Type[Matrix]:
        """Returns the calculated matrix.

        Returns
        -------

        matrix : type(fitmodel.Matrix)
            Calculated Matrix
        """
        raise NotImplementedError

    @abstractmethod
    def estimated_matrix(self) -> Type[Matrix]:
        """Returns the estimated matrix.

        Returns
        -------

        matrix : type(fitmodel.Matrix)
            Estimated Matrix
        """
        raise NotImplementedError

    @abstractmethod
    def dataset_class(self) -> Type[DatasetDescriptor]:
        """Returns an implementation for model.DatasetDescriptor.

        Returns
        -------

        descriptor : type(model.DatasetDescriptor)
            Implementation of model.DatasetDescriptor
        """
        raise NotImplementedError

    @abstractmethod
    def fit_model_class(self) -> Type[FitModel]:
        """Returns an implementation for fitmodel.FitModel.

        Returns
        -------

        fitmodel : type(fitmodel.FitModel)
            Implementation of fitmodel.FitModel
        """
        raise NotImplementedError

    def fit(self, *args, nnls=False, **kwargs) -> Type[Result]:
        """ Fits the model and returns the result.

        Parameters
        ----------
        nnls :
            (Default value = False)
        *args :

        **kwargs :


        Returns
        -------
        result : type(fitmodel.Result)

        """
        if any([dset.dataset is None for _, dset in self.datasets.items()]):
            raise Exception("Model datasets not initialized")
        return self.fit_model().fit(nnls, *args, **kwargs)

    def simulate(self,
                 dataset: str,
                 axis: Dict[str, np.array],
                 parameter=None,
                 noise=False,
                 noise_std_dev=1.0,
                 ):
        """Simulates the model.

        Parameters
        ----------
        dataset : str
            Label of the dataset to simulate

        axis : dict(str, np.array)
            A dictory with axis
        parameter :
            (Default value = None)
        noise :
            (Default value = False)
        noise_std_dev :
            (Default value = 1.0)

        """
        data = self.dataset_class()(dataset)
        sim_parameter = self.parameter.as_parameters_dict().copy()
        if parameter is not None:
            for k, val in parameter.items():
                k = "p_" + k.replace(".", "_")
                sim_parameter[k].value = val
        for label, val in axis.items():
            data.set_axis(label, val)
        self.datasets[dataset].dataset = data

        kwargs = {}
        kwargs['dataset'] = dataset
        kwargs['noise'] = noise
        kwargs['noise_std_dev'] = noise_std_dev
        data = self.fit_model().eval(sim_parameter, **kwargs)
        self.get_dataset(dataset).set(data)

    def concentrations(self, dataset: str) -> np.ndarray:
        """Returns the precited concentrations for a dataset.

        Parameters
        ----------
        dataset : str
            Label of the dataset

        Returns
        -------
        concentrations : numpy.ndarray
        """
        parameter = self.parameter.as_parameters_dict().copy()
        kwargs = {}
        kwargs['dataset'] = dataset
        return self.fit_model().c_matrix(parameter, **kwargs)

    def fit_model(self) -> FitModel:
        """Returns an instance of the models fitmodel.FitModel implementation.

        Returns
        -------

        fitmodel : fitmodel.FitModel
        """
        return self.fit_model_class()(self)

    @property
    def compartments(self) -> List[str]:
        """A list of compartment labels."""
        return self._compartments

    @compartments.setter
    def compartments(self, value):
        self._compartments = value

    @property
    def parameter(self) -> ParameterGroup:
        """The model parameters."""
        return self._parameter

    @parameter.setter
    def parameter(self, val):
        if not isinstance(val, ParameterGroup):
            raise TypeError
        self._parameter = val

    @property
    def megacomplexes(self) -> Dict[str, Megacomplex]:
        """A dictonary of megacomplexes."""
        return self._megacomplexes

    @megacomplexes.setter
    def megacomplexes(self, value):
        if not isinstance(value, dict):
            raise TypeError("Megacomplexes must be dict.")
        if any(not issubclass(type(value[val]), Megacomplex) for val in value):
            raise TypeError("Megacomplexes must be subclass of 'Megacomplex'")
        self._megacomplexes = value

    def add_megacomplex(self, megacomplex: Megacomplex):
        """Adds a megacomplex to the model.

        Parameters
        ----------
        megacomplex : Megacomplex

        """
        if not issubclass(type(megacomplex), Megacomplex):
            raise TypeError("Megacomplexes must be subclass of 'Megacomplex'")
        if self.megacomplexes is not None:
            if megacomplex.label in self.megacomplexes:
                raise Exception("Megacomplex labels must be unique")
            self.megacomplexes[megacomplex.label] = megacomplex
        else:
            self.megacomplexes = {megacomplex.label: megacomplex}

    def data(self) -> Generator[DatasetDescriptor, None, None]:
        """Gets all datasets as a generator.

        Returns
        -------

        datasets : generator(DatasetDescriptor)
        """
        for _, dataset in self.datasets.items():
            yield dataset.data

    def list_datasets(self) -> List[str]:
        """Returns a list of all dataset labels

        Returns
        -------

        datasets : list(str)
        """
        return [label for label in self.datasets]

    @property
    def datasets(self) -> Dict[str, DatasetDescriptor]:
        """A dictonary of all datasets"""
        return self._datasets

    @datasets.setter
    def datasets(self, value):
        if not isinstance(value, OrderedDict):
            raise TypeError("Datasets must be Ordered.")
        if any(not issubclass(type(value[val]),
                              DatasetDescriptor) for val in value):
            raise TypeError("Dataset must be subclass of 'DatasetDescriptor'")
        self._datasets = value

    def add_dataset(self, dataset: DatasetDescriptor):
        """Adds a DatasetDescriptor to the model

        Parameters
        ----------
        dataset : DatasetDescriptor


        """
        if not issubclass(type(dataset), DatasetDescriptor):
            raise TypeError("Dataset must be subclass of 'DatasetDescriptor'")
        self.datasets[dataset.label] = dataset

    def get_dataset(self, label: str) -> Dataset:
        """ Sets the dataset of a DatasetDescriptor

        Parameters
        ----------
        label : str
            Label of the DatasetDescriptor

        Returns
        -------
        dataset : Dataset
            The Dataset


        """
        return self.datasets[label].dataset

    def set_dataset(self, label: str, dataset: Dataset):
        """ Sets the dataset of a DatasetDescriptor

        Parameters
        ----------
        label : str
            Label of the DatasetDescriptor

        dataset : Dataset
            The Dataset


        """
        self.datasets[label].dataset = dataset

    @property
    def initial_concentrations(self) -> Dict[str, InitialConcentration]:
        """A Dictoinary of the initial concentrations."""
        return self._initial_concentrations

    @initial_concentrations.setter
    def initial_concentrations(self, value):
        if not isinstance(value, dict):
            raise TypeError("Initial concentrations must be dict.")
        if any(not isinstance(value[val],
                              InitialConcentration) for val in value):
            raise TypeError("Initial concentrations must be instance of"
                            " 'InitialConcentration'")
        self._initial_concentrations = value

    def add_initial_concentration(self, initial_concentration: InitialConcentration):
        """Adds an initial concentration to the model.

        Parameters
        ----------
        initial_concentration : InitialConcentration


        """
        if not isinstance(initial_concentration, InitialConcentration):
            raise TypeError("Initial concentrations must be instance of"
                            " 'InitialConcentration'")
        if self.initial_concentrations is not None:
            self.initial_concentrations[initial_concentration.label] =\
                initial_concentration
        else:
            self.initial_concentrations = {initial_concentration.label:
                                           initial_concentration}

    def __str__(self):
        string = "# Model\n\n"
        string += "_Type_: {}\n\n".format(self.type_string())

        string += "## Parameter\n{}\n".format(self.parameter)

        if self.datasets is not None:
            string += "\n## Datasets\n\n"

            for data in self._datasets:
                string += "{}\n".format(self._datasets[data])

        if self.compartments is not None:
            string += "\n## Compartments\n\n"

            for c in self.compartments:
                string += f"* {c}\n"

        string += "\n## Megacomplexes\n\n"

        for mcp in self._megacomplexes:
            string += "{}\n".format(self._megacomplexes[mcp])

        if self.initial_concentrations is not None:
            string += "\n## Initital Concentrations\n\n"

            for i in self._initial_concentrations:
                string += "{}\n".format(self._initial_concentrations[i])

        return string



from c_matrix_gaussian_irf import calculateCSingleGaussian
from c_matrix import calculateC


class CMatrixCython(object):
    def c_matrix(self, C, idxs, rates, times, scale):
        return calculateC(C, idxs, rates, times, scale)

    def c_matrix_gaussian_irf(self, C, idxs, rates, times, centers, widths,
                              scale, backsweep, backsweep_period):
        calculateCSingleGaussian(C, idxs, rates, times, centers, widths,
                                 scale, backsweep, backsweep_period)



""" Glotaran Parameter"""

from math import isnan

from lmfit import Parameter as LmParameter


class Parameter(LmParameter):
    """Wrapper for lmfit parameter."""
    def __init__(self):
        self.index = -1
        self.fit = True
        self.label = None
        super(Parameter, self).__init__()

    @classmethod
    def from_parameter(cls, label: str, parameter: LmParameter):
        """Creates a parameter from an lmfit.Parameter

        Parameters
        ----------
        label : str
            Label of the parameter
        parameter : lmfit.Parameter
            lmfit.Parameter
        """
        p = cls()
        p.label = label
        p.value = parameter.value
        p.min = parameter.min
        p.max = parameter.max
        p.vary = parameter.vary
        return p

    @property
    def label(self) -> str:
        """Label of the parameter"""
        return self._label

    @label.setter
    def label(self, label: str):
        self._label = label

    @property
    def fit(self):
        """Whether the paramater should be included in fit. Set false for e.g.
        dormant parameter."""
        return self._fit

    @fit.setter
    def fit(self, value):
        if not isinstance(value, bool):
            raise TypeError("Fit must be True or False")
        self._fit = value

    @LmParameter.value.setter
    def value(self, val):
        if not isinstance(val, (int, float)):
                try:
                    val = float(val)
                except Exception:
                    raise Exception("Parameter Error: value must be numeric:"
                                    "{} Type: {}".format(val, type(val)))

        if isinstance(val, int):
            val = float(val)

        if isnan(val):
            self.vary = False

        LmParameter.value.fset(self, val)

    def __str__(self):
        """ """
        return f"__{self.label}__: _Value_: {self.value}, _Min_:" + \
               f" {self.min}, _Max_: {self.max}, _Vary_: {self.vary}, _Fit_: {self.fit}"



# This is just a placeholder setup.py to claim the glotaran name on PyPI
# It is not meant to be usuable in any way as of yet.
import os
import shutil

from setuptools import setup, find_packages, Command
from setuptools.extension import Extension


class CleanCommand(Command):
    """Custom clean command to tidy up the project root."""
    user_options = [("all", "a", "")]

    def initialize_options(self):
        self.all = True
        self._clean_tree = [
            os.path.abspath("./build"),
            os.path.abspath("./dist"),
        ]
        for root, dirs, filenames in os.walk(os.path.abspath(".")):
            for dir in dirs:
                if dir.endswith(("__pycache__", ".egg-info", ".pytest_cache")) and \
                        ".tox" not in root:

                    self._clean_tree.append(os.path.join(root, dir))

            for filename in filenames:
                if filename.endswith((".pyc", ".pyd", ".tgz", ".egg-info")) and \
                        ".tox" not in root and root not in self._clean_tree:

                    self._clean_tree.append(os.path.join(root, filename))

    def finalize_options(self):
        pass

    def run(self):
        for clean_path in self._clean_tree:
            try:
                shutil.rmtree(clean_path)
            except Exception:
                pass


try:
    import numpy
    import scipy
    ext_modules = [
        Extension("c_matrix",
                  ["glotaran/models/spectral_temporal/c_matrix_cython/c_matrix.pyx"],
                  include_dirs=[numpy.get_include(), scipy.get_include(),
                                "glotaran/models/spectral_temporal/c_matrix_cython"]),
        Extension("c_matrix_gaussian_irf",
                  ["glotaran/models/spectral_temporal/c_matrix_cython/c_matrix_gaussian_irf.pyx",
                   "glotaran/models/spectral_temporal/c_matrix_cython/erfce.c"],
                  include_dirs=[numpy.get_include(), scipy.get_include(),
                                "glotaran/models/spectral_temporal/c_matrix_cython"]),
        #  Extension("c_matrix_damped_oscillation",
        #            ["glotaran/models/damped_oscillation/c_matrix_damped_oscillations.pyx"],
        #            include_dirs=[numpy.get_include(), scipy.get_include(),
        #                          "glotaran/models/damped_oscillation"]),
                  ]

except ImportError:
    raise ImportError("To install glotaran you need to have following packages installed:\n"
                      "numpy>=1.9.1\n"
                      "scipy>=1.0.0\n"
                      "Cython>=0.28.3\n"
                      "You can install them by running:\n"
                      "`pip install 'numpy>=1.9.1' 'scipy>=1.0.0' 'Cython>=0.28.3'`")

setup_requires = [
    'numpy>=1.9.1',
    'scipy>=1.0.0',
    'Cython>=0.28.3',
    'setuptools>=39.2.0'
]
install_requires = [
    'lmfit>=0.9.7',
    'pandas>=0.23.1',
    'pyyaml',
    'matplotlib',  # dependency introduced by glotaran.plotting
    'natsort',  # dependency introduced by glotaran.data.io.chlorospec_format
    'lmfit-varpro>=0.0.1'
]

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="glotaran",
    version='0.0.5',
    description='The Glotaran fitting engine.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='http://glotaran.org',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Cython',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Chemistry',
    ],
    author='Joris Snellenburg, Joern Weissenborn, Ivo van Stokkum',
    author_email="""j.snellenburg@gmail.com,
                    joern.weissenborn@gmail.com,
                    i.h.m.van.stokkum@vu.nl """,
    license='GPLv3',
    python_requires=">=3.6",
    packages=find_packages(),
    setup_requires=setup_requires,
    install_requires=setup_requires+install_requires,
    cmdclass={'clean': CleanCommand},
    ext_modules=ext_modules,
    test_suite='glotaran',
    tests_require=['pytest'],
    zip_safe=False
)

#    package_data={'glotaran.models.kinetic.c_matrix_opencl':
#                  ['*.cl']},



""" Glotaran K-Matrix """

from collections import OrderedDict
from typing import List, Tuple
import numpy as np
import scipy

from glotaran.model import InitialConcentration, ParameterGroup


class KMatrix:
    """ A K-Matrix represents a first order differental system."""
    def __init__(self, label: str, matrix: OrderedDict, compartments: List[str]):
        """

        Parameters
        ----------
        label : str
            Label of the K-Matrix

        matrix : OrderedDict(tuple(str, str), str)
            Dictonary with the matrix entries in 'to-from' notation.

        compartments : list(str)
            A list of all compartments in the model.

        """
        self.label = label
        self.matrix = matrix
        self._create_compartment_map(compartments)
        # We keep track of all compartments to combine later.
        self._all_compartments = compartments

    @classmethod
    def empty(cls, label: str, compartments: List[str]):
        """ Creates an empty K-Matrix. Useful for combining.

        Parameters
        ----------
        label : str
            Label of the K-Matrix

        matrix : OrderedDict(tuple(str, str), str)
            Dictonary with the matrix entries in 'to-from' notation.

        compartments : list(str)
            A list of all compartments in the model.
        """
        return cls(label, OrderedDict(), compartments)

    @property
    def label(self) -> str:
        """ Label of the K-Matrix"""
        return self._label

    @label.setter
    def label(self, value: str):
        self._label = value

    @property
    def matrix(self) -> OrderedDict:
        """ Dictonary with the matrix entries in 'to-from' notation. """
        return self._matrix

    @matrix.setter
    def matrix(self, value: OrderedDict):
        if not isinstance(value, OrderedDict):
            raise TypeError("Matrix must be OrderedDict like {(1,2):value}")
        self._matrix = value

    @property
    def involved_compartments(self) -> List[str]:
        """ A list of all compartments in the matrix. """
        compartments = []
        for index in self.matrix:
            compartments.append(index[0])
            compartments.append(index[1])

        compartments = list(set(compartments))
        return compartments

    @property
    def compartment_map(self):
        """ A list of all compartments in the matrix. Index in the list
        correlates to index in the matrix."""
        return self._compartment_map

    def _create_compartment_map(self, compartments):
        self._compartment_map = [c for c in compartments if c in
                                 self.involved_compartments]

    def combine(self, k_matrix: "KMatrix") -> "KMatrix":
        """ Creates a combined matrix.

        Parameters
        ----------
        k_matrix : KMatrix
            Matrix to combine with.


        Returns
        -------
        combined : KMatrix
            The combined KMatrix.

        """
        #  if isinstance(k_matrix, list):
        #      next = k_matrix[1:] if len(k_matrix) > 2 else k_matrix[1]
        #      return self.combine(k_matrix[0]).combine(next)
        if not isinstance(k_matrix, KMatrix):
            raise TypeError("K-Matrices can only be combined with other"
                            "K-Matrices.")
        combined_matrix = OrderedDict()
        for entry in k_matrix.matrix:
            combined_matrix[entry] = k_matrix.matrix[entry]
        for entry in self.matrix:
            combined_matrix[entry] = self.matrix[entry]
        return KMatrix("{}+{}".format(self.label, k_matrix.label),
                       combined_matrix, self._all_compartments)

    def asarray(self) -> np.ndarray:
        """ Depricated, only used for testing"""
        compartment_map = self.compartment_map
        size = len(compartment_map)
        array = np.zeros((size, size), dtype=np.int32)
        # Matrix is a dict
        for index in self.matrix:
            i = compartment_map.index(index[0])
            j = compartment_map.index(index[1])
            array[i, j] = self.matrix[index]
        return np.array(array, copy=True)

    def full(self, parameter):
        """[ 0 k3
          k1 k2]

        translates to

        [ -k1 k3
          k1 -k2-k3]

        this still correct

        d/dt S1 = -k1 S1 + k3 S2
        d/dt S2 = +k1 S1 - k2 S2 - k3 S2

        it helps to read it so:


        [-k1     k3       [S1]
        k1  -k2-k3]    [S2]

        Parameters
        ----------
        parameter :


        Returns
        -------

        """

        size = len(self.compartment_map)
        mat = np.zeros((size, size), np.float64)
        for (to_comp, from_comp), param in self.matrix.items():
            to_idx = self.compartment_map.index(to_comp)
            fr_idx = self.compartment_map.index(from_comp)
            param = parameter.get(param)

            if to_idx == fr_idx:
                mat[to_idx, fr_idx] -= param
            else:
                mat[to_idx, fr_idx] += param
                mat[fr_idx, fr_idx] -= param
        return mat

    def eigen(self, parameter: ParameterGroup) -> Tuple[np.ndarray,
                                                        np.ndarray]:
        """ Returns the eigenvalues and eigenvectors of the k matrix.

        Parameters
        ----------
        parameter : glotaran.model.ParameterGroup


        Returns
        -------
        (eigenvalues, eigenvectos) : tuple(np.ndarray, np.ndarray)

        """
        matrix = self.full(parameter)
        # get the eigenvectors and values
        eigenvalues, eigenvectors = np.linalg.eig(matrix)
        return (eigenvalues.real, eigenvectors.real)

    def a_matrix(self,
                 initial_concentration: InitialConcentration,
                 parameter: ParameterGroup) -> np.ndarray:
        initial_concentration = initial_concentration.parameter
        initial_concentration = \
            [initial_concentration[self._all_compartments.index(c)] for c in
             self.compartment_map]
        initial_concentration = [parameter.get(i) for i in initial_concentration]

        _, eigenvectors = self.eigen(parameter)
        gamma = np.matmul(scipy.linalg.inv(eigenvectors),
                          initial_concentration)

        a_matrix = np.empty(eigenvectors.shape, dtype=np.float64)

        for i in range(eigenvectors.shape[0]):
            a_matrix[i, :] = eigenvectors[:, i] * gamma[i]

        return a_matrix

    def __str__(self):
        """ """

        longest = max([len(s) for s in self.compartment_map]) + 6
        header = "compartment |"
        if longest < len(header):
            longest = len(header)

        longest_h = max([len(f" __{c}__ |") for c in self.compartment_map]) + 3
        longest_p = max([len(str(self.matrix[i])) for i in self.matrix]) + 3
        if longest_p < longest_h:
            longest_p = longest_h

        string = f"### _{self.label}_\n"
        string += "\n"
        #  string += "```\n"
        #  string += f"\t"
        header_sub = "|".rjust(longest, "-")
        for comp in self.compartment_map:
            header += f" __{comp}__ |".rjust(longest_p)
            header_sub += "|".rjust(longest_p, "-")
        string += header
        string += "\n"
        string += header_sub
        string += "\n"
        for comp in self.compartment_map:
            string += f"__{comp}__ |".rjust(longest)
            for comp2 in self.compartment_map:
                found = False
                for index in self.matrix:
                    if index[0] == comp and index[1] == comp2:
                        found = True
                        string += f" {self.matrix[index]} |".rjust(longest_p)
                if not found:
                    string += " |".rjust(longest_p)
            string += "\n"
        return string



# Glotaran package __init__.py

# workaround to import Model and Dataset with
# `from glotaran import Model`/`from glotaran import Dataset`

from glotaran.specification_parser import parse_yml  # noqa: E402
from glotaran.specification_parser import parse_file  # noqa: E402
from glotaran import dataio, model  # noqa: E402

__version__ = '0.0.5'

Dataset = model.Dataset
DatasetDescriptor = model.DatasetDescriptor
Model = model.Model
ParameterGroup = model.ParameterGroup

# Top level API
# SeparableModel = separable_model.SeparableModel
# SeparableModelResult = result.SeparableModelResult

# loadSomethingFromFile
# Load Glotaran analysis protocol (*.gat or *.yml) from file

load = parse_file  # short for loadProtocol
parse = parse_yml  # short for loadProtocol

io = dataio

# # user calls this object whatever
# glotaran._loadProtocol()
# glotaran._loadProtocolFromFile() #or path
# glotaran._loadProtocolFromString()
#
# # data file vs dataset ??
# glotaran.add_dataset()
# glotaran._add_dataset_from_file() #or path
# glotaran._add_dataset_from_clipboard()
#
# glotaran.read_data_file
# glotaran.read_multiple_data_files
#
# # model specification
#
# glotaran.free_parameter('parameter')
# glotaran.free_parameter_for_dataset(parameter, index)
#
# ## model evaluation
#
# glotaran.fit() #use lmfit to minimize
# evaluate the current model at given timepoints and wavelengths and produce data object
# glotaran.calculate()
#
#
# ## getting results
#
# glotaran.parameters
# glotaran.get_fitted_parameters
#
# glotaran.das
# glotaran.get_decay_associated_spectra
#
# glotaran.eas
# glotaran.get_evolution_associated_spectra
#
# glotaran.sas
# glotaran.get_species_associated_spectra
#
# glotaran.concentrations
# glotaran.get_concentrations
#
# glotaran.residuals
# glotaran.get_residuals
# glotaran.get_residuals_for_dataset(index i)
#
# glotaran.plot_das
# glotaran.plot_eas
# glotaran.plot_sas
# glotaran.plot_concentrations
# glotaran.plot_spectra



from enum import Enum


class SpectralUnit(Enum):
    pixel = 0
    nm = 1
    per_cm = 2


class SpectralTimetrace(object):
    """
    Represents a spectral timetrace
    """
    def __init__(self, spectra, timepoints, spectral_indices=[],
                 spectral_unit=SpectralUnit.pixel, timeunit="s"):
        if not isinstance(spectral_unit, SpectralUnit):
            raise TypeError
        self.spectral_unit = spectral_unit
        if spectral_indices is []:
            for i in range(spectra[0]):
                spectral_indices.append(i)
        self._spectral_indices = spectral_indices
        channel_labels = []
        for i in spectral_indices:
            channel_labels.append(i)

    def wavenumbers(self):
        if self.spectral_unit is SpectralUnit.nm:
            wn = []
            for wl in self.wavelengths():
                wn.append(10000000/wl)
            return wn
        elif self.spectral_unit is SpectralUnit.per_cm:
            return self.spectral_indices
        else:
            raise Exception("Spectral unit is pixel.")

    def wavelengths(self):
        if self.spectral_unit is SpectralUnit.per_cm:
            wl = []
            for wn in self.wavenumbers():
                wl.append(10000000/wn)
            return wl
        elif self.spectral_unit is SpectralUnit.nm:
            return self.spectral_indices
        else:
            raise Exception("Spectral unit is pixel.")



"""Glotaran Parameter Group"""

from typing import Generator, Tuple
from collections import OrderedDict

from lmfit import Parameters

from .parameter import Parameter


class ParameterGroup(OrderedDict):
    """Represents are group of parameters. Can contain other groups, creating a
    tree-like hirachy."""
    def __init__(self, label: str):
        self._label = label
        self._parameters = OrderedDict()
        self._fit = True
        self._root = None
        super(ParameterGroup, self).__init__()

    @classmethod
    def from_parameter_dict(cls, parameter: Parameters):
        """Creates a parameter group from an lmfit.Parameters dictionary

        Parameters
        ----------
        parameter : lmfit.Parameters
            lmfit.Parameters dictionary
        """

        root = cls("p")

        for lbl, param in parameter.items():
            lbl = lbl.split("_")
            if len(lbl) is 2:
                # it is a root param
                param = Parameter.from_parameter(lbl.pop(), param)
                root.add_parameter(param)
                continue

            # remove root
            lbl.pop(0)

            top = root
            while len(lbl) is not 0:
                group = lbl.pop(0)
                if group in top:
                    if len(lbl) is 1:
                        param = Parameter.from_parameter(lbl.pop(), param)
                        top[group].add_parameter(param)
                    else:
                        top = top[group]
                else:
                    group = ParameterGroup(group)
                    top.add_group(group)
                    if len(lbl) is 1:
                        param = Parameter.from_parameter(lbl.pop(), param)
                        group.add_parameter(param)
                    else:
                        top = group
        return root

    def add_parameter(self, parameter: Parameter):
        """

        Parameters
        ----------
        parameter : Parameter

        """
        if not isinstance(parameter, list):
            parameter = [parameter]
        if any(not isinstance(p, Parameter) for p in parameter):
            raise TypeError("Parameter must be  instance of"
                            " glotaran.model.Parameter")
        for p in parameter:
            p.index = len(self._parameters) + 1
            if p.label is None:
                p.label = "{}".format(p.index)
            if not self.fit:
                p.fit = False
            self._parameters[p.label] = p

    def add_group(self, group: 'ParameterGroup'):
        """

        Parameters
        ----------
        group : ParameterGroup

        """
        if not isinstance(group, ParameterGroup):
            raise TypeError("Leave must be glotaran.model.ParameterGroup")
        if not self.fit:
            group.fit = False
        group.set_root(self)
        self[group.label] = group

    def set_root(self, root):
        self._root = root

    def get_nr_roots(self):
        n = 0
        root = self._root
        while root is not None:
            n += 1
            root = root._root
        return n

    @property
    def fit(self):
        """Indicates if the group should be filtered out of the fitting process"""
        return self._fit

    @fit.setter
    def fit(self, value):
        for p in self.all_group():
            p.fit = value
        self._fit = value
        for _, l in self.items():
            l.fit = value

    @property
    def label(self):
        """Label of the group """
        return self._label

    def groups(self) -> Generator['ParameterGroup', None, None]:
        """Generator over all groups and their subgroups"""
        for group in self:
            for l in group.groups():
                yield l

    def get(self, label: str) -> Parameter:
        """Gets a parameter by it label.

        Parameters
        ----------
        label : str
            Label of the Parameter to get.


        Returns
        -------
        parameter : Parameter

        """

        # sometimes the spec parser delivers the labels as int
        label = str(label)

        path = label.split(".")
        label = path.pop()

        group = self
        for l in path:
            group = group[l]
        try:
            return group._parameters[label]
        except KeyError:
            raise Exception("Cannot find parameter "
                            "{}".format(".".join(path)+"."+label))

    def get_by_index(self, idx: int) -> Parameter:
        """ Gets a parameter by its index. Only works for unlabeled parameters
        in the root group.

        Parameters
        ----------
        idx : int
            Index of the parameter.

        Returns
        -------
        parameter : Parameter
        """
        return [i for _, i in self._parameters.items()][idx-1]

    def all_group(self) -> Generator[Parameter, None, None]:
        """Generator returning all Parameter within the group, but not in subgroups"""
        for _, p in self._parameters.items():
            yield p

    def all(self) -> Generator[Parameter, None, None]:
        """Generator returning all parameters within the group and in subgroups"""
        for p in self.all_group():
            yield p
        for l in self:
            for p in self[l].all():
                yield p

    def all_with_label(self, root) -> Generator[Tuple[str, Parameter], None,
                                                None]:
        """ Same as all, but returns the labels relative to the given root
        group.
        Parameters
        ----------
        root : label of the root group


        """
        root = "{}_{}".format(root, self.label) if root is not None else \
            self.label
        for label, p in self._parameters.items():
            yield ("{}_{}".format(root, label), p)
        for _, l in self.items():
            for (lbl, p) in l.all_with_label(root):
                yield (lbl, p)

    def as_parameters_dict(self, only_fit=False) -> Parameters:
        """
        Creates a lmfit.Parameters dict.

        Parameters
        ----------
        only_fit : bool
            (Default value = False)
            if True, all parameters with fit = False will be filtered

        Returns
        -------
        Parameters : lmfit.Parameters
        """
        params = Parameters()
        for (label, p) in self.all_with_label(None):
            p.name = label
            if not only_fit or p.fit:
                params.add(p)
        return params

    def __str__(self):
        t = "".join(["  " for _ in range(self.get_nr_roots())])
        s = ""
        if self.label is not "p":
            s += f"{t}* __{self.label}__:\n"
        for _, p in self._parameters.items():
            s += f"{t}  * {p}\n"
        for _, g in self.items():
            s += f"{g.__str__()}"
        return s



"""Glotaran Dataset"""

from typing import Tuple
from copy import copy
import numpy as np


class Dataset(object):
    """Dataset encapsulates actual data and labeled axis. This class serves as
    base class for classes which wrap e.g. a CSV file.

    Parameters
    ----------
    label : Label of the Datasets

    Returns
    -------
    """

    def __init__(self, label):
        self.label = label
        self._axis = {}

    @property
    def label(self):
        """Label of the dataset """
        return self._label

    @label.setter
    def label(self, label):
        """

        Parameters
        ----------
        label : Label of the Dataset


        Returns
        -------

        """
        self._label = label

    def get_estimated_axis(self):
        """Get the axis along the calculated matrices are grouped"""
        raise NotImplementedError

    def get_calculated_axis(self):
        """Get the axis along the estimated matrices are grouped"""
        raise NotImplementedError

    def get_axis(self, label):
        """

        Parameters
        ----------
        label : label of the axis


        Returns
        -------
        axis: list or np.ndarray
        """
        return self._axis[label]

    def set_axis(self, label, axis):
        """

        Parameters
        ----------
        label : label of the axis

        axis: list or np.ndarray


        Returns
        -------

        """
        if not isinstance(axis, (list, np.ndarray)):
            raise TypeError("Axis must be list or ndarray")
        if any(not _hashable(v) for v in axis):
            raise ValueError("Axis must be list or ndarray of hashable values")
        self._axis[label] = axis

    def get(self) -> np.array:
        """nd.array of shape (M,N)

        Returns
        -------
        data : np.array


        """
        return self._data

    def set(self, data):
        """

        Parameters
        ----------
        data : np.array


        """
        if not isinstance(data, np.ndarray):
            raise TypeError("Data must be a nd array")
        if len(data.shape) is not 2:
            raise ValueError("Dataset must be 2-dimensional")
        self._data = data

    def copy(self) -> 'Dataset':
        """Returns a copy of the Dataset Object

        Returns
        -------
        dataset : Dataset
        """
        return copy(self)

    def svd(self) -> Tuple[np.array, np.array, np.array]:
        """Returns the singular value decomposition of the dataset


        Returns
        -------
        tuple :
            (lsv, svals, rsv)

            lsv : np.array
                left singular values
            svals : np.array
                singular values
            rsv : np.array
                right singular values

        """
        lsv, svals, rsv = np.linalg.svd(self.get().T)
        return lsv, svals, rsv.T


def _hashable(value):
    try:
        # pylint: disable=pointless-statement
        {value: None}
        return True
    except Exception:
        return False



import os
import csv
from ast import literal_eval as make_tuple

from .utils import get_keys_from_object, parse_range_list

from glotaran.model import (EqualAreaConstraint,
                            EqualConstraint,
                            InitialConcentration,
                            Parameter,
                            ParameterGroup,
                            ZeroConstraint,
                            )


class Keys:
    BOUND = "bound"
    COMPARTMENT = "compartment"
    COMPARTMENTS = "compartments"
    COMPARTMENT_CONSTRAINTS = "compartment_constraints"
    COMPARTMENT_SCALING = "compartment_scaling"
    DATASET = "dataset"
    DATASETS = "datasets"
    EQUAL = "equal"
    EQUAL_AREA = "equal_area"
    EQUAL_AREA = "equal_area"
    EXPR = "expr"
    FIT = "fit"
    FIX = "fix"
    INITIAL_CONCENTRATION = "initial_concentration"
    INTERVALS = "intervals"
    LABEL = "label"
    MAX = "max"
    MEGACOMPLEX = "megacomplex"
    MEGACOMPLEXES = "megacomplexes"
    MEGACOMPLEX_SCALING = "megacomplex_scaling"
    MIN = "min"
    PARAMETER = "parameter"
    PARAMETERS = "parameters"
    PARAMETER_CONSTRAINTS = "parameter_constraints"
    PATH = "path"
    RANGE = "range"
    SCALING = "scaling"
    SUBBLOCKS = "sub_blocks"
    TARGET = "target"
    TARGETS = "targets"
    TO = "to"
    TYPE = 'type'
    VALUE = 'value'
    VARY = "vary"
    WEIGHT = "weight"
    ZERO = "zero"


ModelParser = {}


def get_model_parser(spec):
    if spec[Keys.TYPE] in ModelParser:
        return ModelParser[spec[Keys.TYPE]](spec)
    else:
        raise Exception("Unsupported model type {}."
                        .format(spec[Keys.TYPE]))


def register_model_parser(type_name, parser):
    if not issubclass(parser, ModelSpecParser):
        raise TypeError
    ModelParser[type_name] = parser


class ModelSpecParser(object):

    def __init__(self, spec):
        self.spec = spec
        self.model = self.get_model()()

    def get_model(self):
        raise NotImplementedError

    def get_megacomplexes(self):
        raise NotImplementedError

    def get_additionals(self):
        raise NotImplementedError

    def get_dataset_descriptor(self, label, initial_concentration,
                               megacomplexes, megacomplex_scalings,
                               dataset_scaling, compartment_scalings,
                               compartment_constraints):
        raise NotImplementedError

    def get_dataset(self, dataset_spec):
        label = dataset_spec[Keys.LABEL]
        # path = dataset_spec[DatasetKeys.PATH]
        # type = dataset_spec[Keys.TYPE]
        initial_concentration = \
            dataset_spec[Keys.INITIAL_CONCENTRATION] if \
            Keys.INITIAL_CONCENTRATION in dataset_spec else \
            None
        megacomplexes = dataset_spec[Keys.MEGACOMPLEXES]

        dataset_scaling = \
            dataset_spec[Keys.SCALING] if \
            Keys.SCALING in dataset_spec else None

        cmplx_scalings = {}
        if Keys.MEGACOMPLEX_SCALING in dataset_spec:
            for scaling in dataset_spec[Keys.MEGACOMPLEX_SCALING]:
                (cmplx, params) = \
                    get_keys_from_object(scaling, [Keys.MEGACOMPLEX,
                                                   Keys.PARAMETER])
                cmplx_scalings[cmplx] = params

        compartment_scalings = {}
        if Keys.COMPARTMENT_SCALING in dataset_spec:
            for scaling in dataset_spec[Keys.COMPARTMENT_SCALING]:
                (c, param) = get_keys_from_object(scaling, [Keys.COMPARTMENT,
                                                            Keys.PARAMETER])
                compartment_scalings[c] = param

        compartment_constraints = \
            self.get_compartment_constraints(dataset_spec)

        self.model.add_dataset(
            self.get_dataset_descriptor(label,
                                        initial_concentration,
                                        megacomplexes,
                                        cmplx_scalings,
                                        dataset_scaling,
                                        compartment_scalings,
                                        dataset_spec,
                                        compartment_constraints,
                                        ))

    def get_dataset_additionals(self, dataset, dataset_spec):
        raise NotImplementedError

    def get_datasets(self):
        for dataset_spec in self.spec[Keys.DATASETS]:
            self.get_dataset(dataset_spec)

    def get_parameters(self):
        params = self.spec[Keys.PARAMETERS]
        if isinstance(params, str):
            if os.path.isfile(params):
                f = open(params)
                params = f.read()
                f.close
            dialect = csv.Sniffer().sniff(params.splitlines()[0])

            reader = csv.reader(params.splitlines(), dialect)

            params = []
            for row in reader:
                params += [float(e) for e in row]
        self.model.parameter = self.get_leaf("p", params)

    def get_leaf(self, label, items):
        leaf = ParameterGroup(label)
        for item in items:
            if isinstance(item, dict):
                label, items = list(item.items())[0]
                leaf.add_group(self.get_leaf(label, items))
            elif isinstance(item, bool):
                leaf.fit = item
            else:
                leaf.add_parameter(self.get_parameter(item))
        return leaf

    def get_parameter(self, p):
        param = Parameter()

        if not isinstance(p, list):
            param.value = p
            return param

        def retrieve(filt, default):
            tmp = list(filter(filt, p))
            return tmp[0] if tmp else default

        def filt_float(x):
            if isinstance(x, bool):
                return False
            try:
                float(x)
                return True
            except Exception:
                return False

        param.value = retrieve(filt_float, 'nan')
        param.label = retrieve(lambda x: isinstance(x, str) and not
                               x == 'nan', None)
        param.fit = retrieve(lambda x: isinstance(x, bool), True)
        options = retrieve(lambda x: isinstance(x, dict), None)

        if options is not None:
            if Keys.MAX in options:
                param.max = options[Keys.MAX]
            if Keys.MIN in options:
                param.min = options[Keys.MIN]
            if Keys.EXPR in options:
                param.expr = options[Keys.EXPR]
            if Keys.VARY in options:
                param.vary = options[Keys.VARY]
        return param

    def get_parameter_constraints(self):
        if Keys.PARAMETER_CONSTRAINTS not in self.spec:
            return
        for constraint in self.spec[Keys.PARAMETER_CONSTRAINTS]:
            (type,) = get_keys_from_object(constraint, [Keys.TYPE])

            if type == Keys.FIX:
                (value, params) = get_keys_from_object(constraint,
                                                       [Keys.VALUE,
                                                        Keys.PARAMETERS],
                                                       start=1)

                if isinstance(params, list):
                    params = ", ".join(["{}".format(p) for p in params])

                for p in parse_range_list(params):
                    self.model.parameter.get(p).vary = not value

            elif type == Keys.BOUND:
                (min, max, params) = get_keys_from_object(constraint,
                                                          [Keys.MIN, Keys.MAX,
                                                           Keys.PARAMETERS],
                                                          start=1)
                if isinstance(params, list):
                    params = ", ".join(["{}".format(p) for p in params])
                for p in parse_range_list(params):
                    if max is not None:
                        self.model.parameter.get(p).max = float(max)
                    if min is not None:
                        self.model.parameter.get(p).min = float(min)

    def get_compartment_constraints(self, dataset_spec):
        if Keys.COMPARTMENT_CONSTRAINTS not in dataset_spec:
            return []
        constraints = []
        for constraint in dataset_spec[Keys.COMPARTMENT_CONSTRAINTS]:
            (tpe, c, intv) = get_keys_from_object(constraint,
                                                  [Keys.TYPE,
                                                   Keys.COMPARTMENT,
                                                   Keys.INTERVALS])
            intervals = []
            for interval in intv:
                intervals.append(make_tuple(interval))
            if tpe == Keys.ZERO:
                constraints.append(ZeroConstraint(c, intervals))

            else:
                if tpe == Keys.EQUAL:
                    (target, param) = get_keys_from_object(constraint,
                                                           [Keys.TARGETS,
                                                            Keys.PARAMETERS],
                                                           start=3)
                    constraints.append(
                        EqualConstraint(c, intervals, target, param))
                elif tpe == Keys.EQUAL_AREA:
                    (target, param) = get_keys_from_object(constraint,
                                                           [Keys.TARGET,
                                                            Keys.PARAMETER],
                                                           start=3)
                    (weight,) = get_keys_from_object(constraint,
                                                     [Keys.WEIGHT],
                                                     start=5)
                    constraints.append(
                        EqualAreaConstraint(c, intervals, target,
                                            param, weight))
        return constraints

    def get_initial_concentrations(self):
        if Keys.INITIAL_CONCENTRATION not in self.spec:
            return
        for concentration in self.spec[Keys.INITIAL_CONCENTRATION]:
            (label, parameter) = get_keys_from_object(concentration,
                                                      [Keys.LABEL,
                                                       Keys.PARAMETER])
            self.model.add_initial_concentration(
                InitialConcentration(label, parameter))

    def get_compartments(self):
        self.model.compartments = self.spec[Keys.COMPARTMENTS]

    def parse(self):
        self.get_compartments()
        self.get_parameters()
        self.get_parameter_constraints()
        self.get_initial_concentrations()
        self.get_megacomplexes()
        self.get_datasets()
        self.get_additionals()



from .spectral_shape import SpectralShape


class SpectralShapeGaussian(SpectralShape):
    def __init__(self, label, amplitude, location, width):
        super(SpectralShapeGaussian, self).__init__(label)
        self.amplitude = amplitude
        self.location = location
        self.width = width

    def __str__(self):
        string = super(SpectralShapeGaussian, self).__str__()
        string += ", _Type_: Gaussian"
        string += f", _Amplitude_: {self.amplitude}"
        string += f", _Location_: {self.location}"
        string += f", _Width_: {self.width}"
        return string



""" Glotaran Measured IRF"""

import numpy as np

from .irf import Irf


class MeasuredIrf(Irf):
    """A measured IRF."""

    def __init__(self, label):
        self._data = None
        super(MeasuredIrf, self).__init__(label)

    @property
    def data(self) -> np.ndarray:
        """Measured data."""
        if self._data is None:
            raise Exception(f"{self.label}: data not loaded")
        return self._data

    @data.setter
    def data(self, value: np.ndarray):
        self._data = value

    def type_string(self):
        raise "Measured"

    def __str__(self):
        return f"### _{self.label}_\n* _Type_: {self.type_string()}\n"
